# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A collection of functions and options for the command line interface for soprano."""


__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "August 10, 2022"


import logging
import os
import re
from collections import OrderedDict, defaultdict
from configparser import ConfigParser
from pathlib import Path
from typing import List, Optional

import click
import numpy as np
import pandas as pd
from ase import Atoms
from ase.visualize import view as aseview

from soprano.calculate.nmr import DEFAULT_MARKER_SIZE
from soprano.calculate.nmr.simpson import SimpsonTemplates
from soprano.data.nmr import _el_iso
from soprano.properties.linkage import ElementPairs, Molecules
from soprano.utils import average_quaternions

# join home and config file
home = os.path.expanduser("~")
# get default soprano config file:
DEFAULT_CFG = os.environ.get("SOPRANO_CONFIG", f"{home}/.soprano/config.ini")
DEFAULT_PRECISION = 4

VALID_DF_OUTPUT_EXTENSIONS = ["csv", "json", "html", "md", "tex", "txt", "tsv", "dat", "xls", "xlsx"]


import os

def _validate_df_output_extension(output_filename: str, output_format: Optional[str] = None) -> str:
    """Validate the output filename extension.
    Args:
        output_filename (str): The output filename.
        output_format (str): The output format.
    Returns:
        str: The output filename.
    Raises:
        click.UsageError: If the output format is not valid.
    """
    if output_format is None:
        # infer the output format from the filename extension
        _, ext = os.path.splitext(output_filename)
        inferred_format = ext[1:]
        output_format = inferred_format
    # check if the output format is valid
    if output_format not in VALID_DF_OUTPUT_EXTENSIONS:
        raise click.UsageError(
            f"Output format must be one of {VALID_DF_OUTPUT_EXTENSIONS}. "
            f"Got: {ext[1:]} (inferred from filename: {output_filename})."
        )
    return output_format

# callback to load config file
def configure(ctx, param, filename):
    cfg = ConfigParser()
    cfg.read(filename)
    ctx.default_map = {}
    for sect in cfg.sections():
        command_path = sect.split(".")
        if command_path[0] != "soprano":
            continue
        defaults = ctx.default_map
        for cmdname in command_path[1:]:
            defaults = defaults.setdefault(cmdname, {})
        defaults.update(cfg[sect])


# We'll need to rename the columns to include units before printing
UNITS = {
    "MS_shielding": "ppm",
    "MS_shift": "ppm",
    "MS_anisotropy": "ppm",
    "MS_reduced_anisotropy": "ppm",
    "MS_span": "ppm",
    "MS_alpha": "deg",
    "MS_beta": "deg",
    "MS_gamma": "deg",
    "EFG_Vzz": "Vm^-2",
    "EFG_quadrupolar_constant": "MHz",
    "EFG_alpha": "deg",
    "EFG_beta": "deg",
    "EFG_gamma": "deg",
    "EFG_NQR": "MHz",
    "D": "kHz",
    "alpha": "deg",
    "beta": "deg",
    "gamma": "deg",
}


# TODO: write guide for this on website...
NO_CIF_LABEL_WARNING = """
## Protip: ##
This .magres file doesn't seem to have CIF-stlye labels.
Using these is considered a good idea, but is not required.
You can export these automatically from a cif file using 
cif2cell. e.g. for CASTEP:

cif2cell mystructure.cif --export-cif-labels -p castep

"""

# Function to expand aliases
def expand_aliases(input_list, alias_dict):
    if input_list is None:
        return None
    output_list = []
    for item in input_list:
        if item in alias_dict:
            output_list.extend(alias_dict[item])
        else:
            output_list.append(item)
    return output_list

### PARSER HELPERS ###
def isotope_selection(ctx, parameter, isotope_string):
    """Parse the isotope string.
    Args:
        ctx: click context
        parameter: click parameter
        isotope_string (str): The isotopes specification, in the form ``"2H,15N" for deuterium and 15N``.
    Returns:
        dict: The isotope for each element specified. Formatted as::
            {Element: Isotope}.

    """
    if isotope_string == "":
        return {}
    isotope_dict = {}
    for sym in isotope_string.split(","):
        # make sure there are only integers and A-z in the string
        if not re.match(r"^[A-z0-9]+$", sym):
            raise click.BadParameter("Isotope string must be of the form '2H,15N'")
        try:
            el, isotope = _el_iso(sym)
        except Exception as e:
            raise e
        isotope_dict[el] = isotope
    return isotope_dict


def keyvalue_parser(ctx, parameter, value):
    """Parse strings in the form 'C:1,H:2' into a dictionary.
        Also accept = as the separator between key and value.
        e.g. the MS shift reference and gradient strings.
    Args:
        ctx: click context
        parameter: click parameter
        value (str): The references specification, in the form ``"C:100,H:123"``.
                     If value is a single float, that will returned instead of a dict.
    Returns:
        dict: The values for each key specified. Formatted as::
            {key: value}.
    """

    keyvaluedict = {}
    if value != "":
        for sym in re.split(",", value):
            try:
                el, reference = re.split(":|=", sym)
                keyvaluedict[el] = float(reference)
            except Exception as e:
                raise e
    return keyvaluedict


def get_column_list(ctx, parameter, value):
    """Parse the column names string.
    TODO: Document this a bit better.
    Args:
        ctx: click context
        parameter: click parameter
        value (str): The column names, comma-separated.
                    Some shortcuts defined for MS_angles and EFG_angles.
    Returns:
        list: The column names specified.
    """
    if value == "all" or value is None:
        return None
    specified_columns = [c.strip() for c in value.split(",")]
    # make sure no duplicates, preserving order
    specified_columns = list(OrderedDict.fromkeys(specified_columns))
    return specified_columns


### CLI OPTIONS ###
config = click.option(
    "-c",
    "--config",
    type=click.Path(dir_okay=False),
    default=DEFAULT_CFG,
    callback=configure,
    is_eager=True,
    expose_value=False,
    show_default=True,
    help="Read option defaults from the specified INI file"
    "If not set, first checks environment variable: "
    "``SOPRANO_CONFIG`` and then ``~/.soprano/config.ini``",
)

ms_tag = click.option(
    "--ms-tag",
    default="ms",
    help="Name of array containing MS tensors. default: 'ms'",
)
efg_tag = click.option(
    "--efg-tag", 
    default="efg",
    help="Name of array containing EFG tensors. default: 'efg'",
)

subset_help = """Selection string of subset of sites include. e.g. 
                ``-s C`` for only and all carbon atoms,
                ``-s C.1-3,H.1.2`` for carbons 1,2,3 and hydrogens 1 and 2,
                ``-s C1,H1a,H1b`` for any sites with the labels C1, H1a and H1b."""
# option to select a subset of atoms
subset = click.option(
    "--subset",
    "-s",
    "subset",
    type=str,
    default=None,
    help=subset_help,
)
#  what to extract/analyse
nmrproperties = click.option(
    "--properties",
    "-p",
    type=click.Choice(["efg", "ms"]),
    default=["efg", "ms"],
    multiple=True,
    help="Properties for which to extract and summarise e.g. ``-p ms.`` "
    "They can be combined by using the flag multiple times: ``-p ms -p efg.`` "
    "Defaults to both ms and efg.",
)
df_output = click.option(
    "--output",
    "-o",
    type=click.Path(exists=False, writable=True, dir_okay=False),
    default=None,
    help="Output file name. If not specified, output is printed to stdout."
    "If the output file name has a recognised extension (see ``--output-format`` for allowed values), the format is guessed from that."
    "The format can be overridden with the ``--output-format`` option.",
)
df_output_format = click.option(
    "--output-format",
    "-f",
    default=None,
    type=click.Choice(VALID_DF_OUTPUT_EXTENSIONS),
    help="Output file format. "
    "If not specified, the format is guessed from output filename extension."
    "If specified, it overrides the output filename extension."
    "The ``txt``, ``tsv`` and ``dat`` formats all produce tab separated files."
    "The ``md`` format produces a markdown table and requires an optional dependency on the tabulate package.",
)
# merge output files
df_merge = click.option(
    "--merge",
    "-m",
    is_flag=True,
    default=False,
    help="If present, merge all files into a single output file.",
)

# Option to specify the isotopes to use
isotopes = click.option(
    "--isotopes",
    "-i",
    callback=isotope_selection,
    default="",
    metavar="ISOTOPES",
    help="Isotopes specification (e.g. ``-i 13C`` for carbon 13 "
    "``-i 2H,15N`` for deuterium and 15N). "
    "When nothing is specified it defaults to the most common NMR active isotope.",
)
# flag option to reduce by symmetry
reduce = click.option(
    "--reduce/--no-reduce",
    is_flag=True,
    default=True,
    help="Reduce the output by symmetry-equivalent sites. "
    "If there are CIF-style labels present, then these override the symmetry-grouping in "
    "case of a clash (though there should never be a clash...). "
    "Note that this doesn't take into account magnetic symmetry! "
    "Defaults to True, so use ``--no-reduce`` to turn off symmetry reduction.",
)
# symprec flag
symprec = click.option(
    "--symprec",
    type=click.FLOAT,
    default=1e-4,
    help="Symmetry precision for symmetry reduction. " "Defaults to 1e-4.",
)
# option to specify group_pattern for averaging
average_group = click.option(
    "--average-group",
    "-g",
    type=str,
    default=None,
    help="Group pattern for averaging. "
    "Currently only works for XHn groups such as CH3, CH2, NH2 etc. "
    "You can specify several, comma separated as in ``-g CH3,CH2,NH2``. "
    "If not specified, no averaging is performed.",
)

# optional argument for the precision of the output
precision = click.option(
    "--precision",
    type=click.INT,
    default=3,
    help="Precision of the output (number of decimal places). Defaults to 3.",
)
# choose between Euler angle conventions 'zyz' or 'zxz'
euler = click.option(
    "--euler",
    "euler_convention",
    type=click.Choice(["zyz", "zxz"]),
    default="zyz",
    help="Convention for Euler angles. Defaults to ``zyz``.",
)
# sort by df column
df_sortby = click.option(
    "--sortby",
    type=str,
    default=None,
    help="Sort by column. Defaults to sorting by site number. "
    "It can be any column in the output. "
    "For example ``--sortby EFG_Vzz``",
)
df_sort_order = click.option(
    "--sort-order",
    type=click.Choice(["ascending", "descending"]),
    default="ascending",
    help="Sort order. Defaults to ascending.",
)
# dictionary option to specify reference for each element
references = click.option(
    "--references",
    callback=keyvalue_parser,
    default="",
    help="Reference shielding for each element (in ppm). "
    "The format is ``--references C:170,H:123``. ",
)
gradients = click.option(
    "--gradients",
    callback=keyvalue_parser,
    default="",
    help="Reference shielding gradients for each element. "
    "Defaults to -1 for all elements. Set it like this: "
    "``--gradients H:-1,C:-0.97``. ",
)
# TODO: have an option to set a file/env variable for the references...
# flag to include certain columns only
df_include = click.option(
    "--include",
    callback=get_column_list,
    default='minimal',
    help="Include only certain columns. "
    "The columns are specified as a comma-separated list. "
    "For example ``--include MS_shielding,EFG_Vzz``. "
    "Defaults to all columns.",
)
# flag to exclude certain columns
df_exclude = click.option(
    "--exclude",
    callback=get_column_list,
    default=None,
    help="Exclude certain columns. "
    "The columns are specified as a comma-separated list. "
    "For example ``--exclude MS_alpha,MS_beta,MS_gamma``. "
    "Defaults to None.",
)
# flag to filter results
df_query = click.option(
    "--query",
    type=str,
    default=None,
    help="Filter results based on query. "
    "The filter is specified as a pandas query. "
    "Note that you must enclose the query in quotes! "
    "Refer to the column names without the units.  "
    "For example ``--query 'MS_shielding > 100'``. "
    "You can combine queries with ``and`` and ``or`` etc. "
    "e.g. ``--query 'MS_shielding > 100 and MS_shielding < 180'``. "
    "Defaults to no filter.",
)

# flag to view
view = click.option(
    "--view",
    is_flag=True,
    default=False,
    help="If present, view the structure(s) with the ASE GUI."
    "Note that the ASE GUI can color the sites according to their tags. "
    "This can be used to see what sites were tagged as equivalent.",
)


# verbosity flag e.g. -v -vv -vvv
verbosity = click.option(
    "--verbosity",
    "-v",
    count=True,
    help="Increase verbosity. "
    "Use -v for info, -vv for debug. Defaults to only showing warnings and errors.",
)


## plotting options
# plot type argument
plot_type = click.option(
    "--plot_type", "-p", type=click.Choice(["2D", "1D"]), default="2D", help="Plot type"
)
# x-element
plot_xelement = click.option(
    "--xelement",
    "-x",
    "x_element",
    type=str,
    required=True,
)
# y element
plot_yelement = click.option(
    "--yelement",
    "-y",
    "y_element",
    type=str,
    required=False,
    help="Element to plot on the y-axis. "
    "If not specified, but a 2D plot is requested, the x-element is used.",
)
# rcut
plot_rcut = click.option(
    "--rcut",
    "rcut",
    type=float,
    default=10,
    help="Cutoff distance for plotting. Defaults to 10 Angstrom.",
)
plot_yaxis_order = click.option(
    "--yaxis-order",
    "yaxis_order",
    type=click.Choice(["1Q", "2Q"]),
    default="1Q",
    help="Single (1Q) or double (2Q) quantum order for the y-axis. Defaults to 1Q.",
)
plot_showdiagonal = click.option(
    "--diag/--no-diag",
    "show_diagonal",
    default=True,
    help="Plot diagonal line if the x and y elements are the same. "
    "Defaults to True (i.e. showing the diagonal line).",
)
plot_showgrid = click.option(
    "--grid/--no-grid",
    "show_grid",
    default=True,
    help="Show grid lines at the axis ticks. "
    "Defaults to True (i.e. showing the grid lines).",
)
plot_show_ticklabels = click.option(
    "--ticklabels/--no-ticklabels",
    "show_ticklabels",
    default=True,
    help="Show tick labels. " "Defaults to True (i.e. showing the tick labels).",
)
plot_showconnecors = click.option(
    "--connectors/--no-connectors",
    "show_connectors",
    default=True,
    help="Show horizontal connectors between points in the case of a 2Q plot. "
    "Defaults to True (i.e. show connectors).",
)

# flip x and y
# plot_flipx = click.option('--flipx',
#                 '-fx',
#                 'flip_x',
#                 is_flag=True,
#                 default=False,
#                 help='Flip x axis')
# plot_flipy = click.option('--flipy',
#                 '-fy',
#                 'flip_y',
#                 is_flag=True,
#                 default=False,
#                 help='Flip y axis')
# # x-axis label
# plot_xlabel = click.option('--xlabel',
#                 type=str,
#                 default='',
#                 help='Custom X-axis label')
# # y-axis label
# plot_ylabel = click.option('--ylabel',
#                 type=str,
#                 default='',
#                 help='Custom Y-axis label.')
# x-axis range
plot_xlims = click.option(
    "--xlim",
    nargs=2,
    type=float,
    default=None,
    help="X-axis range. For example ``--xlim 20 100``",
)
# y-axis range
plot_ylims = click.option(
    "--ylim",
    nargs=2,
    type=float,
    default=None,
    help="Y-axis range. For example ``--ylim 20 100``",
)
plot_show_markers = click.option(
    "--markers/--no-markers",
    "show_markers",
    default=True,
    help="Show markers? Default is True.",
)

# marker
plot_marker = click.option(
    "--marker-symbol",
    type=str,
    default="+",
    help="Marker type. "
    "For example ``--marker-symbol o``. "
    "Accepts any matplotlib marker type.",
)

# marker color
plot_marker_color = click.option(
    "--marker-color",
    "marker_color",
    type=str,
    default='C0',
    help="Marker color. Default is 'C0'.",
)

# scale marker size by value
plot_scale_marker_by = click.option(
    "--scale-marker-by",
    type=click.Choice(["fixed", "distance", "inversedistance", "dipolar", "jcoupling"]),
    default="fixed",
    help="Scale marker size by chosen property. "
    "``fixed`` means that all the markers will have the same size. "
    "``distance`` means that the marker size will be proportional to the distance between the sites. "
    "``inversedistance`` means that the marker size will be proportional to the inverse of the distance between the sites. "
    "``dipolar`` means that the marker size will be proportional to the dipolar coupling between the sites. "
    "Default is ``fixed``.",
)
# marker size
plot_max_marker_size = click.option(
    "--max-marker-size",
    "-ms",
    type=float,
    default=DEFAULT_MARKER_SIZE,
    help=f"Maximum marker size. Default is {DEFAULT_MARKER_SIZE}.",
)

# marker-line width
plot_marker_linewidth = click.option(
    "--marker-linewidth",
    type=float,
    default=0.5,
    help="Marker linewidth. Default is 0.5.",
)

# show marker legend?
plot_marker_legend = click.option(
    "--legend/--no-legend",
    "show_marker_legend",
    default=False,
    help="Show marker legend? Default is True.",
)
# show heatmap?
plot_show_heatmap = click.option(
    "--heatmap/--no-heatmap",
    "show_heatmap",
    default=False,
    help="Show heatmap? Default is False.",
)

# x broadening - None means default to 5% of the range. Otherwise float in ppm
plot_xbroadening = click.option(
    "--xbroadening",
    type=float,
    default=None,
    help="Broadening of the x-axis in ppm. "
    "Defaults to 5% of the range. "
    "Set to 0 to turn off broadening.",
)

# y broadening - None means default to 5% of the range. Otherwise float in ppm
plot_ybroadening = click.option(
    "--ybroadening",
    type=float,
    default=None,
    help="Broadening of the y-axis in ppm. "
    "Defaults to 5% of the range. "
    "Set to 0 to turn off broadening.",
)

# colour map
plot_colormap = click.option(
    "--colormap",
    "-cmap",
    type=str,
    default="bone",
    help="Colour map for the heatmap. Default is 'bone'. "
    "See https://matplotlib.org/stable/tutorials/colors/colormaps.html for more options. "
    "Try adding '_r' to the end of the colormap name to reverse it.",
)
# show contour?
plot_show_contour = click.option(
    "--contour/--no-contour",
    "show_contour",
    default=False,
    help="Show contour? Default is False.",
)

# contour levels
plot_contour_levels = click.option(
    "--contour-levels",
    type=int,
    default=10,
    help="Number of contour levels. Default is 10.",
)

# contour color
plot_contour_color = click.option(
    "--contour-color",
    type=str,
    default="C1",
    help="Contour color. Default is 'C1'.",
)

# contour linewidth
plot_contour_linewidth = click.option(
    "--contour-linewidth",
    type=float,
    default=0.2,
    help="Contour linewidth. Default is 0.5.",
)


# plot filename
plot_output = click.option(
    "--output",
    "-o",
    "plot_filename",
    type=click.Path(exists=False, dir_okay=False, writable=True),
    required=False,
    help="Name of the plot file. "
    "If not specified, the plot will be displayed in a window. "
    "The file extension determines the plot type (svg or pdf recommended).",
)
plot_shielding = click.option(
    "--shielding",
    "plot_shielding",
    is_flag=True,
    default=None,
    help="Force plot shielding (even if references are present). "
    "Default is to plot shifts if references are given but shielding if no references given).",
)


# option to select a subset of atoms
dip_selection_i = click.option(
    "--select_i", "-i", "selection_i", type=str, default=None, help=subset_help
)
dip_selection_j = click.option(
    "--select_j", "-j", "selection_j", type=str, default=None, help=subset_help
)
dip_rss_flag = click.option(
    "--rss",
    "rss_flag",
    is_flag=True,
    default=False,
    help="Calculate the dipolar constant Root Sum Square for each atom in a system (includes periodicity)."
    "(As opposed to the dipolar couplings between pairs of atoms)."
    "Default is False.",
)
dip_rss_cutoff = click.option(
    "--cutoff",
    "rss_cutoff",
    type=float,
    default=5.0,
    help="Cutoff for Dipolar constant Root Sum Square for each atom in a system (includes periodicity)."
    " Units are Angstroms. Default is 5.0 Angstroms.",
)
dip_isonuclear = click.option(
    "--isonuclear",
    is_flag=True,
    default=False,
    help="Include only dipolar interactions between atoms of the same type."
    "Default is False.",
)


#### Groups of CLI options
# options that apply to pandas dataframes
DF_OPTIONS = [
    df_output_format,
    df_output,
    df_merge,
    df_sortby,
    df_sort_order,
    df_include,
    df_exclude,
    df_query,
]

NMR_OPTIONS = [
    ms_tag,
    efg_tag,
    nmrproperties,
    isotopes,
    average_group,
    reduce,
    euler,
    references,
    gradients,
    subset,
]

COMMON_OPTIONS = [
    config,
    verbosity,
    view,
    symprec,
    precision,
]
PLOT_SPECIFIC_OPTIONS = [
    isotopes,
    average_group,
    euler,
    references,
    gradients,
    subset,
    reduce,
    df_query,
    plot_type,
    plot_xelement,
    plot_yelement,
    plot_rcut,
    plot_yaxis_order,
    # plot_flipx,
    # plot_flipy,
    # plot_xlabel,
    # plot_ylabel,
    plot_xlims,
    plot_ylims,
    plot_show_markers,
    plot_marker,
    plot_max_marker_size,
    plot_marker_linewidth,
    plot_scale_marker_by,
    plot_marker_color,
    plot_marker_legend,
    plot_showdiagonal,
    plot_showgrid,
    plot_showconnecors,
    plot_show_ticklabels,
    plot_show_heatmap,
    plot_xbroadening,
    plot_ybroadening,
    plot_colormap,
    plot_show_contour,
    plot_contour_levels,
    plot_contour_color,
    plot_contour_linewidth,
    plot_output,
    plot_shielding,
]

DIP_OPTIONS = [
    isotopes,
    average_group,
    dip_selection_i,
    dip_selection_j,
    dip_rss_flag,
    dip_rss_cutoff,
    dip_isonuclear,
]

NMREXTRACT_OPTIONS = COMMON_OPTIONS + NMR_OPTIONS + DF_OPTIONS
DIPOLAR_OPTIONS = COMMON_OPTIONS + DIP_OPTIONS + DF_OPTIONS
PLOT_OPTIONS = COMMON_OPTIONS + PLOT_SPECIFIC_OPTIONS
VIEW_OPTIONS = [config, average_group, subset, reduce, symprec]


# function to add options to a subcommand
def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


### FUNCTIONS ###
def viewimages(images, reload_as_molecular=None):
    """
    Use ASE GUI to view the images.

    If they contain C and H, we'll assume it's a molecular
    crystal and reload it as such.

    Args:
        images (list): list of ASE Atoms objects to view
        reload_as_molecular (bool): whether to reload as a molecular crystal. Default is None. This will be set to True if the images contain a C-H.

    """
    if reload_as_molecular is None:
        images = [reload_as_molecular_crystal(image) for image in images]
    elif reload_as_molecular:
        # reload as molecular crystal True -> force
        images = [reload_as_molecular_crystal(image, force=True) for image in images]
    else:
        pass

    aseview(images)


def print_results(
    dfs, output=None, output_format="csv", precision=DEFAULT_PRECISION, verbose=True
):
    # set pandas print precision
    pd.set_option("display.precision", precision)
    # make sure we output all rows, even if there are lots!
    pd.set_option("display.max_rows", 999)
    nframes = len(dfs)
    if output:
        output_format = _validate_df_output_extension(output, output_format)
        for i, df in enumerate(dfs):
            if nframes > 1:
                # then we want to write out
                # each dataframe to a separate file
                # so let's prefix the filename
                magrespath = df["file"].iloc[0]
                # Use pathlib for cross-platform path handling
                path_parts = Path(magrespath).stem.split("/")
                prefix = "_".join(path_parts) + "-"
            else:
                prefix = ""
            fname = prefix + output
            if verbose:
                click.echo(f"Writing output to {fname}")

            if output_format == "csv":
                df.to_csv(fname, index=True)
            elif output_format == "json":
                df.to_json(fname)
            elif output_format == "html":
                df.to_html(fname)
            elif output_format == "tex":
                df.to_latex(fname)
            elif output_format == "md":
                # this requires the tabulate package
                # I'm not sure if it's worth adding as a dependency...
                df.to_markdown(fname)
            elif (
                output_format == "txt"
                or output_format == "tsv"
                or output_format == "dat"
            ):
                df.to_csv(fname, index=True, sep="\t")
            elif output_format == "xls" or output_format == "xlsx":
                df.to_excel(fname, index=True)
            else:
                raise ValueError(f"Unknown output format: {output_format}")
    else:
        for df in dfs:
            # Replace NaN values with "-" for display purposes
            display_df = df.replace(pd.NaT, "-").replace(pd.NA, "-").fillna("-")

            if nframes == 1 and df["file"].nunique() > 1:
                click.echo(display_df)
            else:
                fname = df["file"].iloc[0]
                click.echo(f"\n\nExtracted data from: {fname}")
                # Drop the 'file' column for display
                display_df.drop("file", axis=1, inplace=True)
                click.echo(display_df)


def units_rename(colname, units_dict=UNITS):
    for key, unit in units_dict.items():
        if key in colname:
            return f"{colname}/{unit}"
    # if no matches found, return original name
    return colname


def apply_df_filtering(
    df: pd.DataFrame,
    include: List,
    exclude: List,
    query: str,
    essential_columns: List = [],
    logger: Optional[logging.Logger] = None,
) -> pd.DataFrame:
    """
    Inlcude/exclude columns and filter the dataframe using a pandas query.

    Args:
        df (pd.DataFrame): the dataframe to filter
        include (list): list of columns to include
        exclude (list): list of columns to exclude
        query (str): pandas query string to filter the dataframe
        essential_columns (list): list of columns that must be included
        logger (logging.Logger): logger to use

    Returns:
        pd.DataFrame: the filtered dataframe

    """

    # if no logger specified, use the default
    if not logger:
        logger = logging.getLogger(__name__)

    if query:
        # use pandas query to filter the dataframe
        logger.info(f"\nFiltering dataframe using query: {query}")
        df.query(query, inplace=True)
        logger.info(f"-----> Filtered to {len(df)} sites.")

    # what columns should we include/exclude?

    if include:
        # what columns should we include/exclude?
        specified_columns = [c for c in include if c not in essential_columns]
        logger.debug(f"\nIncluding only columns containing: {specified_columns}")
        columns_to_include = essential_columns + specified_columns
        missing_columns = get_missing_cols(df, columns_to_include)
        if len(missing_columns) > 0:
            logger.warning(
                f"These columns specified {missing_columns}"
                f" do not match any in the dataframe ({df.columns})"
            )
        columns_to_include = get_matching_cols(df, columns_to_include)
        df = df[columns_to_include].copy()
    if exclude:
        logger.debug(f"\nExcluding columns: {exclude}")
        # remove those that are already not in df
        specified_columns = get_matching_cols(df, exclude)
        df = df.drop(specified_columns, axis=1)
    # drop any that have only NaN values
    df = df.dropna(axis=1, how="all")
    return df


def sortdf(df, sortby, sort_order):
    """sort df by column, return new df"""

    if sortby:
        if sortby in df.columns:
            ascending = sort_order == "ascending"
            if sortby == "labels":
                isalpha = df[sortby].str.isalpha()
                if all(isalpha):
                    # no cif style labels, so sort by alphabetical order
                    df = df.sort_values(by=sortby, ascending=ascending)
                elif any(isalpha):
                    # tricky -- we have a mix of cif style labels and non-cif style labels
                    # so we need to sort by the cif style labels first, then sort by the non-cif style labels
                    # this is a bit of a hack, but it works
                    raise NotImplementedError(
                        "Sorting by labels is not implemented"
                        "for a mix of cif and non-cif -style labels"
                    )
                else:
                    # all cif stlye
                    # some care is needed to sort by labels in a natural way
                    df[["_str", "_int"]] = df[sortby].str.extract(r"([a-zA-Z]*)(\d*)")
                    df["_int"] = df["_int"].astype(int)

                    df = df.sort_values(by=["_str", "_int"], ascending=ascending).drop(
                        ["_int", "_str"], axis=1
                    )
            else:
                df = df.sort_values(by=sortby, ascending=ascending)
        else:
            raise ValueError(f"{sortby} not found in summary columns names")
    return df


def get_matching_cols(df, lst):
    """
    Get the columns of a dataframe that roughly match a list of strings.
    """
    return [col for col in df.columns if any(x in col for x in lst)]


def get_missing_cols(df, lst):
    """
    Get the items in list that don't match any of the columns of a dataframe
    """
    return [x for x in lst if all(x not in col for col in df.columns)]


def get_duplicates(seq):
    """
    Returns dict {duplicate_value: [indices]} for duplicates in a list
    """
    tally = defaultdict(list)
    for i, item in enumerate(seq):
        tally[item].append(i)
    return dict([(key, locs) for key, locs in tally.items() if len(locs) > 1])


def average_quaternions_by_tags(quaternions, tags):
    """
    For repeated tags, average the quaternions.
    Return the modified list of quaternions.
    """
    if len(set(tags)) != len(tags):
        dupl_dict = get_duplicates(tags)
        for tag, idx in dupl_dict.items():
            quat_group = [quaternions[ig] for ig in idx]
            # compute group average quaternion
            quat_av = average_quaternions(quat_group)
            # update list of quaternions
            for ig in idx:
                quaternions[ig] = quat_av
    return quaternions


def find_XHn_groups(atoms, pattern_string, tags=None, vdw_scale=1.0):
    """Find groups of atoms based on a functional group pattern.
    The pattern is a string such as CH3 or CH2.
    It must contain an element symbol, H and the number of H atoms


    | Args:
    |   atoms (ase.Atoms): Atoms object on which to perform selection
    |   pattern_string (str): functional group pattern e.g. 'CH3'
    |                        for a methyl group. Assumes the group is
    |                        the thing(s) connected to the first atom.
    |                        They can be combined, comma separated.
    |                        TODO: add SMILES/SMARTS support?
    |   vdw_scale (float): scale factor for vdw radius (used for bond searching)
    """
    from soprano.properties.linkage import Bonds

    if tags is None:
        tags = np.arange(len(atoms))

    bcalc = Bonds(vdw_scale=vdw_scale, return_matrix=True)
    bonds, bmat = bcalc(atoms)
    all_groups = []
    for group_pattern in pattern_string.split(","):
        # split into central element and number of H atoms
        if "H" not in group_pattern:
            raise ValueError(
                f"{group_pattern} is not a valid group pattern "
                "(must contain an element symbol, H, and the number of H atoms. e.g. CH3)"
            )
        X, n = group_pattern.split("H")
        n = int(n)
        # Find XHn groups
        symbs = np.array(atoms.get_chemical_symbols())
        hinds = np.where(symbs == "H")[0]
        groups = []
        xinds = np.where(symbs == X)[0]
        xinds = xinds[np.where(np.sum(bmat[xinds][:, hinds], axis=1) == n)[0]]
        # group_tags = np.ones((len(xinds), n), dtype=int)
        seen_tags = []
        for ix, xind in enumerate(xinds):
            bonded_hinds = np.where(bmat[xind][hinds] == 1)[0]
            group = list(hinds[bonded_hinds])
            assert len(group) == n
            match = []
            if len(seen_tags) > 0:
                match = np.where((seen_tags == tags[group]).all(axis=1))[0]

            if len(match) == 1:
                # how to handle this?
                groups[match[0]] += group
            elif len(match) == 0:
                seen_tags.append(tags[group])
                groups.append(group)
            else:
                raise ValueError(f"Found multiple matches for {group_pattern}")

        all_groups.append(groups)

    return all_groups


def reload_as_molecular_crystal(atoms: Atoms, force=False) -> Atoms:
    """
    If the atoms object is a molecular crystal, reload it with the correct
    connectivity.

    We check if there are any C-H bonds, and if so, we assume it's a molecular
    crystal and reload it as such. If there are no C-H bonds, we assume it's
    not and return the original atoms object, unless force=True.


    Args:
        atoms (ASE Atoms object): the atoms object to be reloaded.
        force (bool): if True, force the reload even if there are no C-H bonds.

    Returns:
        atoms (ASE Atoms object): the atoms object with the correct connectivity.
    """
    # save initial order
    atoms.set_array("order_tag", np.arange(len(atoms)))
    if has_CH_bonds(atoms) or force:
        # temporarily translate the atoms to the COM
        com = atoms.cell.T.dot([0.5, 0.5, 0.5]) - atoms.get_center_of_mass()
        atoms.translate(com)
        # let's assume this is an organic molecule/crystal
        # and try to reload the atoms object with the correct
        # connectivity:
        mols = Molecules.get(atoms)
        if len(mols) == 0:
            return atoms

        temp = mols[0].subset(atoms, use_cell_indices=True)
        if len(mols) > 1:
            for mol in mols[1:]:
                temp.extend(mol.subset(atoms, use_cell_indices=True))
        # restore original order
        temp = temp[temp.get_array("order_tag").argsort()]
        # restore original centering
        temp.translate(-com)
        atoms = temp

    return atoms


def has_CH_bonds(atoms: Atoms, rcut: float = 1.5) -> bool:
    """Check if the atoms object has bonds with CH atoms
    Does this simply by comparing the labels with the element symbols.
    If they're all the same, then so special labels are present.
    """
    # First check if it contains C and H elements
    symbols = atoms.get_chemical_symbols()
    if not ("C" in symbols and "H" in symbols):
        return False

    # Test the return_pairs=False option
    dists = ElementPairs.get(
        atoms, element1="C", element2="H", return_pairs=False, rcut=rcut
    )
    return len(dists) > 0



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Module containing AtomsProperties that relate to linkage properties of a given
system, self-correlation etc.
"""


from soprano.properties.linkage.linkage import (
    BondGraph,
    Bonds,
    CoordinationHistogram,
    DihedralAngleList,
    ElementPairs,
    HydrogenBonds,
    HydrogenBondsNumber,
    LinkageList,
    MoleculeCOM,
    MoleculeCOMLinkage,
    MoleculeMass,
    MoleculeNumber,
    MoleculeQuaternion,
    MoleculeRelativeRotation,
    Molecules,
    MoleculeSpectralSort,
)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Helper functions for NMR calculations and plotting
"""

import itertools
import logging
import re
from collections.abc import Iterable
from dataclasses import dataclass
from functools import wraps
from importlib.resources import files
from typing import Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
from ase import Atoms
from scipy.optimize import minimize

from soprano.properties.labeling.labeling import MagresViewLabels
from soprano.properties.nmr.dipolar import DipolarCoupling
from soprano.properties.nmr.isc import JCIsotropy
from soprano.utils import has_cif_labels


def get_force_matrix(
            positions: np.ndarray,
            positions_original: np.ndarray,
            C:float = 0.01,
            k:float = 0.00001):
    '''
    This is a vectorised version of the above function, which is much faster

    Parameters
    ----------
    positions : np.array
        The y coordinates of the annotations
    positions_original : np.array
        The original y coordinates of the annotations
    C : float, optional
        The repulsive force constant between annotations, by default 0.01
    k : float, optional
        The spring force constant for the spring holding the annotation to it's original position, by default 0.00001

    '''
    Fmat = np.zeros((len(positions), len(positions)))
    # force from other annotations -- ignore the diagonal elements
    displacement_matrix = positions[:, np.newaxis] - positions[np.newaxis, :]
    diag_mask = np.eye(displacement_matrix.shape[0],dtype=bool)
    off_diag_mask = ~diag_mask

    # for any non-zero displacements, calculate the repulsive force
    non_zero_displacements = np.abs(displacement_matrix) > 1e-8
    off_diag_mask = np.logical_and(off_diag_mask, non_zero_displacements)
    # Fmat[off_diag_mask][non_zero_displacements] = -C * displacement_matrix[off_diag_mask][non_zero_displacements] / (np.abs(displacement_matrix[off_diag_mask][non_zero_displacements]))**3

    Fmat[off_diag_mask] += C * displacement_matrix[off_diag_mask] / (np.abs(displacement_matrix[off_diag_mask]))**4

    # if any off-diagonal elements are zero, then set a random force
    zero_off_diag_mask = np.logical_and(off_diag_mask, ~non_zero_displacements)
    # count the number of zero displacements
    if np.sum(zero_off_diag_mask) > 0:
        Fmat[zero_off_diag_mask] = np.random.randn(np.sum(zero_off_diag_mask)) * C * (1/0.05**2)
    # spring force from original position
    Fmat[diag_mask] = -2*k * (positions - positions_original)
    return Fmat

def get_total_forces(
            positions: np.ndarray,
            positions_original: np.ndarray,
            C:float = 0.01,
            k:float = 0.00001):
    '''
    Parameters
    ----------
    positions : np.array
        The y coordinates of the annotations
    positions_original : np.array
        The original y coordinates of the annotations
    C : float, optional
        The repulsive force constant between annotations, by default 0.01
    k : float, optional
        The spring force constant for the spring holding the annotation to it's original position, by default 0.00001
    '''
    Fmat = get_force_matrix(positions, positions_original, C, k)
    # sum over the columns to get the total force on each annotation
    return np.sum(Fmat, axis=1)

def get_total_forces(
            positions: np.ndarray,
            positions_original: np.ndarray,
            C:float = 0.01,
            k:float = 0.00001):
    '''
    Parameters
    ----------
    positions : np.array
        The y coordinates of the annotations
    positions_original : np.array
        The original y coordinates of the annotations
    C : float, optional
        The repulsive force constant between annotations, by default 0.01
    k : float, optional
        The spring force constant for the spring holding the annotation to it's original position, by default 0.00001
    '''
    sign = -1 # trial and error to get the right sign wrt the energy..
    Fmat = sign * get_force_matrix(positions, positions_original, C, k)
    # sum over the columns to get the total force on each annotation
    return np.sum(Fmat, axis=1)
def get_energy(positions, positions_original, C, k):
    '''
    Compute the total energy of the system

    '''
    # compute the energy
    # coulomb energy is the sum of the coulomb energy between each pair of annotations
    # spring energy is the sum of the spring energy for each annotation
    coulomb_energy = 0
    for i in range(len(positions)):
        for j in range(i+1, len(positions)):
            if np.abs(positions[i] - positions[j]) > 1e-8:
                coulomb_energy += C / np.abs(positions[i] - positions[j])**2



    spring_energy = 0.5 * k * np.sum((positions - positions_original)**2)
    # print(f'coulomb_energy: {coulomb_energy}, spring_energy: {spring_energy}')
    energy = spring_energy + coulomb_energy
    return energy


def optimise_annotations(positions, max_iters=10000, C=0.01, k=0.00001, ftol = 1e-3):
    '''
    Parameters
    ----------
    positions : list or np.array
        The x or y coordinates of the annotations to be optimised
    max_iters : int, optional
        The maximum number of iterations to run for, by default 10000
    C : float, optional
        The repulsive force constant between annotations, by default 0.01
    k : float, optional
        The spring force constant for the spring holding the annotation to it's original position, by default 0.00001
    ftol : float, optional
        The tolerance for the forces, by default 1e-3. If all the net forces are below this value, the optimisation will stop
        even if the maximum number of iterations has not been reached.

    '''
    if len(positions) < 3:
        return positions
    # convert to numpy array
    positions = np.array(positions)
    # store original order
    order =  np.argsort(positions)
    # invert order
    inverted_order = np.argsort(order)

    # sort positions
    positions = positions[order]
    ## deep copy of the sorted positions
    positions_original = positions.copy()

    # normalise the positions to the range of the original positions
    original_range = np.max(positions_original) - np.min(positions_original)
    min_pos = np.min(positions_original)
    positions = (positions - np.min(positions_original)) / original_range
    positions_original = (positions_original - min_pos) / original_range

    res = minimize(
            get_energy,
            positions,
            args=(positions_original, C, k),
            jac = get_total_forces,
            # method='BFGS',
            options={'disp': False, 'maxiter':max_iters},
            tol=ftol,
            # add a constraint to keep the first and last annotation in the same place
            constraints = [
                {'type': 'eq', 'fun': lambda x: x[0] - positions_original[0]},
                {'type': 'eq', 'fun': lambda x: x[-1] - positions_original[-1]},
                # another set of constraints to keep all values between 0 and 1 in fractional space
                {'type': 'ineq', 'fun': lambda x: x},
                {'type': 'ineq', 'fun': lambda x: 1-x},
            ]
        )
    new_pos_frac = res.x

    # convert back to original range
    new_pos = new_pos_frac * original_range + min_pos
    positions = new_pos





    # return to original order
    positions = positions[inverted_order]
    return positions



@dataclass
class Peak2D:
    '''
    Class to hold peak data. This is used to store the peak data for a correlation peak in a 2D NMR spectrum.

    The data stored includes:
    - the peak position in x and y
    - the correlation strength
    - the x and y labels for the peak
    - the color of the peak
    - the index of the x and y labels in the list of all labels

    '''
    x: float
    y: float
    xlabel: str
    ylabel: str
    correlation_strength: float = 1
    color: str = 'C0'
    idx_x: Optional[int] = None
    idx_y: Optional[int] = None

    def __repr__(self):
        return f'Peak({self.x}, {self.y}, {self.correlation_strength}, {self.xlabel}, {self.ylabel}, {self.color})'

    def equivalent_to(self, other, xtol=0.005, ytol=0.005, corr_tol=0.1, ignore_correlation_strength=False):
        '''
        Check if two peaks are equivalent. We compare the x and y coordinates and the correlation strength.

        Args:
            other (Peak2D): The other peak to compare to
            xtol (float, optional): The tolerance for the x coordinate. Defaults to 0.005.
            ytol (float, optional): The tolerance for the y coordinate. Defaults to 0.005.
            corr_tol (float, optional): The tolerance for the correlation strength. Defaults to 0.1.
        
        Returns:
            bool: True if the peaks are equivalent, False otherwise
        '''
        x_match = np.abs(self.x - other.x) < xtol
        y_match = np.abs(self.y - other.y) < ytol

        if ignore_correlation_strength:
            corr_match = True
        else:
            corr_match = np.abs(self.correlation_strength - other.correlation_strength) < corr_tol

        return x_match and y_match and corr_match




def lorentzian(X: np.ndarray, x0: float, Y: np.ndarray, y0: float, x_broadening: float, y_broadening: float) -> np.ndarray:
    """
    Calculate the Lorentzian broadening function.

    .. math::
        f(x, y) = \\frac{1}{((x - x_0) / w_x)^2 + ((y - y_0) / w_y)^2 + 1}

    where :math:`w_x` and :math:`w_y` are the broadening factors in the x and y directions, respectively. These
    correspond to the half-width at half-maximum (HWHM) of the Lorentzian function.

    Args:
        X (np.ndarray): Array of x values.
        x0 (float): x-coordinate of the peak.
        Y (np.ndarray): Array of y values.
        y0 (float): y-coordinate of the peak.
        x_broadening (float): Broadening factor in the x direction.
        y_broadening (float): Broadening factor in the y direction.

    Returns:
        np.ndarray: Array of intensity values.
    """
    return np.exp(-((X - x0)**2 / (2 * x_broadening**2) + (Y - y0)**2 / (2 * y_broadening**2)))

def gaussian(X: np.ndarray, x0: float, Y: np.ndarray, y0: float, x_broadening: float, y_broadening: float) -> np.ndarray:
    """
    Calculate the Gaussian broadening function.

    .. math::
        f(x, y) = \\exp\\left(-\\frac{(x - x_0)^2}{2 w_x^2} - \\frac{(y - y_0)^2}{2 w_y^2}\\right)

    where :math:`w_x` and :math:`w_y` are the broadening factors in the x and y directions, respectively. These
    correspond to the standard deviation of the Gaussian function.

    Args:
        X (np.ndarray): Array of x values.
        x0 (float): x-coordinate of the peak.
        Y (np.ndarray): Array of y values.
        y0 (float): y-coordinate of the peak.
        x_broadening (float): Broadening factor in the x direction.
        y_broadening (float): Broadening factor in the y direction.

    Returns:
        np.ndarray: Array of intensity values.
    """
    return np.exp(-((X - x0)**2 / (2 * x_broadening**2) + (Y - y0)**2 / (2 * y_broadening**2)))

def generate_contour_map(
    peaks: List[Peak2D],
    grid_size: int = 100,
    broadening: str = 'gaussian',
    x_broadening: float = 1.0,
    y_broadening: float = 1.0,
    xlims: Optional[Tuple[float, float]] = None,
    ylims: Optional[Tuple[float, float]] = None
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Generate a contour map based on the provided peaks and broadening parameters.

    Args:
        peaks (List[Peak2D]): List of Peak2D objects containing x, y coordinates and correlation strength.
        grid_size (int, optional): Size of the grid for the contour map. Default is 100.
        broadening (str, optional): Type of broadening function to use ('lorentzian' or 'gaussian'). Default is 'lorentzian'.
        x_broadening (float, optional): Broadening factor in the x direction. Default is 1.0.
        y_broadening (float, optional): Broadening factor in the y direction. Default is 1.0.
        xlims (Optional[Tuple[float, float]], optional): Limits for the x-axis. Default is None.
        ylims (Optional[Tuple[float, float]], optional): Limits for the y-axis. Default is None.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Meshgrid arrays X, Y and the intensity grid Z.
    """
    broadening = broadening.lower()
    # Create a grid of x, y values
    if xlims is None:
        x_min, x_max = min(peak.x for peak in peaks), max(peak.x for peak in peaks)
    else:
        x_min, x_max = xlims
    if ylims is None:
        y_min, y_max = min(peak.y for peak in peaks), max(peak.y for peak in peaks)
    else:
        y_min, y_max = ylims

    # Create a grid of x, y values, broadening the grid by 5 times the broadening factor
    x = np.linspace(x_min - 5 * x_broadening, x_max + 5 * x_broadening, grid_size)
    y = np.linspace(y_min - 5 * y_broadening, y_max + 5 * y_broadening, grid_size)
    X, Y = np.meshgrid(x, y)

    # Initialize the intensity grid
    Z = np.zeros_like(X)

    # Apply broadening for each peak, adding to the intensity grid
    for peak in peaks:
        x0, y0, strength = peak.x, peak.y, peak.correlation_strength
        if broadening == 'lorentzian':
            Z += strength * lorentzian(X, x0, Y, y0, x_broadening, y_broadening)
        elif broadening == 'gaussian':
            Z += strength * gaussian(X, x0, Y, y0, x_broadening, y_broadening)
        else:
            raise ValueError(f'Unknown broadening function: {broadening}')

    return X, Y, Z







nmr_base_style = str(files("soprano.calculate.nmr").joinpath("soprano_nmr_base.mplstyle"))
nmr_2D_style = str(files("soprano.calculate.nmr").joinpath("soprano_nmr_2D.mplstyle"))

def styled_plot(*style_sheets):
    """Return a decorator that will apply matplotlib style sheets to a plot.
    ``style_sheets`` is a base set of styles, which will be ignored if
    ``no_base_style`` is set in the decorated function arguments.
    The style will further be overwritten by any styles in the ``style``
    optional argument of the decorated function.
    Args:
        style_sheets (:obj:`list`, :obj:`str`, or :obj:`dict`): Any matplotlib
            supported definition of a style sheet. Can be a list of style of
            style sheets.
    """

    def decorator(get_plot):
        @wraps(get_plot)
        def wrapper(*args, fonts=None, style=None, no_base_style=False, **kwargs):

            if no_base_style:
                list_style = []
            else:
                list_style = list(style_sheets)

            if style is not None:
                if isinstance(style, list):
                    list_style += style
                else:
                    list_style += [style]

            if fonts is not None:
                list_style += [{"font.family": "sans-serif", "font.sans-serif": fonts}]

            plt.style.use(list_style)
            return get_plot(*args, **kwargs)

        return wrapper

    return decorator


def get_atom_labels(atoms: Atoms, logger: Optional[logging.Logger] = None) -> np.ndarray:
    """
    Get the labels for the atoms in the Atoms object. If the labels are not present, they will be generated
    using the MagresViewLabels class.

    Args:
        atoms (ase.Atoms): Atoms object.
        logger (Optional[logging.Logger], optional): Logger object. Default is None.

    Returns:
        np.ndarray: Array of labels.
    """
    if logger is None:
        logger = logging.getLogger(__name__)

    if has_cif_labels(atoms):
        labels = atoms.get_array('labels')
    elif atoms.has('MagresView_labels'):
        # we might have already generated the MV style labels
        labels = atoms.get_array('MagresView_labels')
    else:
        logger.info('No labels found in the atoms object. Generating MagresView-style labels from scratch.')
        labels = MagresViewLabels.get(atoms)
        # convert to numpy array
        labels = np.array(labels, dtype='U25')
    return labels

def prepare_species_labels(isotope, element):
    species_template = r'$\mathrm{^{%s}{%s}}$'
    return species_template % (isotope, element)


def extract_indices(atoms, xelement, yelement):
    idx_x = np.array([atom.index for atom in atoms if atom.symbol == xelement])
    idx_y = np.array([atom.index for atom in atoms if atom.symbol == yelement])

    return idx_x, idx_y


def validate_elements(atoms, xelement, yelement):
    all_elements = atoms.get_chemical_symbols()
    if xelement not in all_elements:
        raise ValueError(f'{xelement} not found in the file after the user-specified filters have been applied.')
    if yelement not in all_elements:
        raise ValueError(f'{yelement} not found in the file after the user-specified filters have been applied.')


def generate_peaks(
    data: List[float],
    pairs: Iterable[Tuple[int, int]],
    labels: List[str],
    markersizes: Union[List[float], float],
    yaxis_order: str,
    xelement: str,
    yelement: str,
) -> List[Peak2D]:
    """
    Generate peaks for the NMR data.

    Args:
        data (List[float]): The NMR data.
        pairs (Iterable[Tuple[int, int]]): Pairs of indices for the peaks.
        labels (List[str]): Labels for the peaks.
        markersizes (Union[List[float], float]): Marker sizes for the peaks.
        yaxis_order (str): Order of the y-axis.
        xelement (str): Element symbol for the x-axis.
        yelement (str): Element symbol for the y-axis.

    Returns:
        List[Peak2D]: List of generated peaks.
    """
    peaks = []
    is_single_marker = isinstance(markersizes, float)
    for ipair, (idx_x, idx_y) in enumerate(pairs):

        x = data[idx_x]
        y = data[idx_y]
        strength = markersizes if is_single_marker else markersizes[ipair]
        xlabel = labels[idx_x]
        ylabel = labels[idx_y]

        if yaxis_order == '2Q':
            y += x
            if xelement == yelement:
                # then we might have both e.g. H1 + H2 and H2 + H1
                # let's set them both to be H1 + H2 by sorting the labels
                xlabel, ylabel = sorted([xlabel, ylabel])
            ylabel = f'{xlabel} + {ylabel}'
        peak = Peak2D(
            x=x,
            y=y,
            correlation_strength=strength,
            xlabel=xlabel,
            ylabel=ylabel,
            idx_x=idx_x,
            idx_y=idx_y)
        peaks.append(peak)
    return peaks

def merge_peaks(
    peaks: Iterable[Peak2D],
    xtol: float = 1e-5,
    ytol: float = 1e-5,
    corr_tol: float = 1e-5,
    ignore_correlation_strength: bool = False
) -> List[Peak2D]:
    """
    Merge peaks that are identical.

    Args:
        peaks (Iterable[Peak2D]): List of peaks to merge.
        xtol (float): Tolerance for x-axis comparison.
        ytol (float): Tolerance for y-axis comparison.
        corr_tol (float): Tolerance for correlation strength comparison.
        ignore_correlation_strength (bool): Whether to ignore correlation strength in comparison.

    Returns:
        List[Peak2D]: List of unique merged peaks.
    """
    # first, get the unique peaks
    unique_peaks = []
    unique_xlabels: Dict[str, Set[str]] = {}
    unique_ylabels: Dict[str, Set[str]] = {}

    peak_map: Dict[Tuple[float, float, float], Peak2D] = {}

    for peak in peaks:
        key = (
            int(peak.x / xtol),
            int(peak.y / ytol),
            int(peak.correlation_strength / corr_tol) if not ignore_correlation_strength else 0
        )
        if key in peak_map:
            unique_peak = peak_map[key]
            unique_xlabels[unique_peak.xlabel].add(peak.xlabel)
            unique_ylabels[unique_peak.ylabel].add(peak.ylabel)
        else:
            unique_peaks.append(peak)
            peak_map[key] = peak
            unique_xlabels[peak.xlabel] = {peak.xlabel}
            unique_ylabels[peak.ylabel] = {peak.ylabel}

    def sort_func(x: str) -> Union[int, str]:
        int_list = re.findall(r'\d+', x)
        return int(int_list[0]) if int_list else x

    for unique_peak in unique_peaks:
        xlabel_list = sorted(unique_xlabels[unique_peak.xlabel], key=sort_func)
        ylabel_list = sorted(unique_ylabels[unique_peak.ylabel], key=sort_func)
        unique_peak.xlabel = '/'.join(xlabel_list)
        unique_peak.ylabel = '/'.join(ylabel_list)

    return unique_peaks

def sort_peaks(peaks: List[Peak2D], priority: str = 'x', reverse: bool=False) -> List[Peak2D]:
    '''
    Sort the peaks in the order of increasing x and y values
    TODO: This doesn't work in some cases. Investigate why.
    '''
    if priority == 'x':
        peaks = sorted(peaks, key=lambda x: (x.x, x.y), reverse=reverse)
    elif priority == 'y':
        peaks = sorted(peaks, key=lambda x: (x.y, x.x), reverse=reverse)
    else:
        raise ValueError(f"Unknown priority: {priority}")
    return peaks


def get_pair_dipolar_couplings(
        atoms: Atoms,
        pairs: Iterable[Tuple[int, int]],
        isotopes: Optional[dict[str, int]] = None,
        unit: str = 'kHz'
    ) -> List[float]:
    """
    Get the dipolar couplings for a list of pairs of atoms.
    For pairs where i == j, the dipolar coupling is set to zero.
    
    Parameters
    ----------
    atoms : ASE Atoms object
        The atoms object that contains the atoms.
    pairs : list of tuples
        List of pairs of atom indices.
    isotopes : dict, optional
    
    Returns
    -------
    dipolar_couplings : list of float
        List of dipolar couplings for the pairs.
    """
    if isotopes is None:
        isotopes = {}

    # Define conversion factors
    conversion_factors = {
        'Hz': 1.0,
        'kHz': 1e-3,
        'MHz': 1e-6
    }

    if unit not in conversion_factors:
        raise ValueError(f"Unsupported unit: {unit}. Supported units are: {', '.join(conversion_factors.keys())}")

    conversion_factor = conversion_factors[unit]

    dipolar_couplings = []
    # This is an explicit loop because we need to get the J coupling for each pair
    # - these might not be the same as the set of pairs that we would get by combining indices
    #   {i} and {j} from the pairs list
    for i, j in pairs:
        if i == j:
            # set the dipolar coupling to zero for pairs where i == j
            dipolar_couplings.append(0)
        else:
            coupling = list(DipolarCoupling.get(atoms, sel_i=[i], sel_j=[j], isotopes=isotopes).values())[0][0]
            dipolar_couplings.append(coupling * conversion_factor)


    return dipolar_couplings

def get_pair_j_couplings(
        atoms: Atoms,
        pairs: Iterable[Tuple[int, int]],
        isotopes: Optional[dict[str, int]] = None,
        unit: str = 'Hz',
        tag: str = 'isc'
    ) -> List[float]:
    """
    Get the J couplings for a list of pairs of atoms.
    For pairs where i == j, the J coupling is set to zero.

    Parameters
    ----------
    atoms : ASE Atoms object
        The atoms object that contains the atoms.
    pairs : list of tuples
        List of pairs of atom indices.
    isotopes : dict, optional
        Dictionary of isotopes for the atoms.
    unit : str, optional
        Unit of the J coupling. Default is 'Hz'.
    tag : str, optional
        Name of the J coupling component to return. Default is 'isc'. Magres files
        usually contain isc, isc_spin, isc_fc, isc_orbital_p and isc_orbital_d.

    Returns
    -------
    j_couplings : list of float
        List of J couplings for the pairs.
    """
    if isotopes is None:
        isotopes = {}

    # Define conversion factors
    conversion_factors = {
        'Hz': 1.0,
        'kHz': 1e-3,
        'MHz': 1e-6
    }

    if unit not in conversion_factors:
        raise ValueError(f"Unsupported unit: {unit}. Supported units are: {', '.join(conversion_factors.keys())}")

    conversion_factor = conversion_factors[unit]

    j_couplings = []
    # This is an explicit loop because we need to get the J coupling for each pair
    # - these might not be the same as the set of pairs that we would get by combining indices
    #   {i} and {j} from the pairs list
    for i, j in pairs:
        if i == j:
            # set the J coupling to zero for pairs where i == j
            j_couplings.append(0)
        else:
            coupling = list(JCIsotropy.get(atoms, sel_i=[i], sel_j=[j], tag=tag, isotopes=isotopes).values())[0]
            j_couplings.append(coupling * conversion_factor)

    return j_couplings



def process_pairs(
    idx_x: np.ndarray,
    idx_y: np.ndarray,
    pairs: Optional[List[Tuple[int, int]]]
) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]], np.ndarray, np.ndarray]:
    """
    Process pairs of indices and update element indices.

    Args:
        idx_x (np.ndarray): Array of x element indices.
        idx_y (np.ndarray): Array of y element indices.
        pairs (Optional[List[Tuple[int, int]]]): List of tuples representing pairs of indices, or None.

    Returns:
        Tuple[List[Tuple[int, int]], List[Tuple[int, int]], np.ndarray, np.ndarray]:
            - List of pairs of indices.
            - List of pairs of element indices.
            - Updated array of x element indices.
            - Updated array of y element indices.
    """
    if pairs:
        pairs_el_idx = []
        for pair in pairs:
            xidx = np.where(idx_x == pair[0])[0][0]  # there should only be one match
            yidx = np.where(idx_y == pair[1])[0][0]  # there should only be one match
            pairs_el_idx.append((xidx, yidx))
    else:
        pairs_el_idx = list(itertools.product(range(len(idx_x)), range(len(idx_y))))
        pairs = list(itertools.product(idx_x, idx_y))

    # Update the indices
    idx_x = np.array(pairs)[:, 0]
    idx_y = np.array(pairs)[:, 1]
    return pairs, pairs_el_idx, idx_x, idx_y

def calculate_distances(pairs: List[Tuple[int, int]], atoms: Atoms) -> np.ndarray:
    """
    Calculate the distances between pairs of atoms.

    Args:
        pairs (List[Tuple[int, int]]): List of tuples representing pairs of atom indices.
        atoms (Atoms): Atoms object that provides the get_distance method.

    Returns:
        np.ndarray: Array of distances corresponding to the pairs.
    """
    pair_distances = np.zeros(len(pairs))
    for i, pair in enumerate(pairs):
        if pair[0] == pair[1]:
            pair_distances[i] = 0.0
        else:
            pair_distances[i] = atoms.get_distance(*pair, mic=True)
    return pair_distances


def filter_pairs_by_distance(
    pairs: List[Tuple[int, int]],
    pairs_el_idx: List[Tuple[int, int]],
    pair_distances: np.ndarray,
    rcut: float
) -> Tuple[List[Tuple[int, int]], List[Tuple[int, int]], np.ndarray, np.ndarray, np.ndarray]:
    """
    Filters pairs of indices based on a cutoff distance.

    Args:
        pairs (List[Tuple[int, int]]): List of tuples representing pairs of indices.
        pairs_el_idx (List[Tuple[int, int]]): List of tuples representing pairs of element indices.
        pair_distances (np.ndarray): Array of distances corresponding to the pairs.
        rcut (float): Cutoff distance for filtering pairs.

    Returns:
        Tuple[List[Tuple[int, int]], List[Tuple[int, int]], np.ndarray, np.ndarray, np.ndarray]:
            - Filtered list of pairs.
            - Filtered list of element index pairs.
            - Filtered array of pair distances.
            - Unique sorted array of x indices.
            - Unique sorted array of y indices.

    Raises:
        ValueError: If no pairs are found after filtering by distance.
    """
    dist_mask = np.where(pair_distances <= rcut)[0]
    pairs_el_idx = [pairs_el_idx[i] for i in dist_mask]
    pairs = [pairs[i] for i in dist_mask]
    pair_distances = pair_distances[dist_mask]

    idx_x = np.unique([pair[0] for pair in pairs])
    idx_y = np.unique([pair[1] for pair in pairs])
    if len(idx_x) == 0 or len(idx_y) == 0:
        raise ValueError('No pairs found after filtering by distance. Try increasing the cutoff distance (rcut).')

    idx_x = np.sort(idx_x)
    idx_y = np.sort(idx_y)

    return pairs, pairs_el_idx, pair_distances, idx_x, idx_y



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Click command line interface for Soprano."""


__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "June 21, 2023"


import logging

import click
import click_log

from soprano.scripts import dipolar, molecules, nmr, nmr_plot, view
# logging
logging.captureWarnings(True)
logger = logging.getLogger(__name__)
click_log.basic_config(logger)

epilog = f"""
    Author: {__author__} ({__email__})\n
    Last updated: {__date__}"""
help_text = """
A CLI tool to streamline common soprano tasks. It has various 
subcommands, each of which has its own set of options and help.
"""


@click.group(
    name="Soprano Command Line Interface",
    help=help_text,
    epilog=epilog,
    invoke_without_command=True,
)
@click_log.simple_verbosity_option(logger)
def soprano():
    pass

soprano.add_command(nmr.nmr)
soprano.add_command(nmr_plot.plotnmr)
soprano.add_command(dipolar.dipolar)
soprano.add_command(view.view)
soprano.add_command(molecules.splitmols)

if __name__ == "__main__":
    soprano()



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Convenience util used to start/stop submitter processes"""



def submitter_handler():

    import argparse as ap
    import os
    import subprocess as sp
    import sys

    from soprano.hpc.submitter import Submitter
    from soprano.utils import import_module

    # First of all, check that we are on Linux, this only works there
    if sys.platform not in ("linux", "linux2"):
        sys.exit("Invalid OS: background Submitter launching works only on " "Linux")

    parser = ap.ArgumentParser(
        description="Use this script to start or stop "
        "a Submitter object as a background process."
    )
    # Required arguments
    parser.add_argument(
        "action",
        type=str,
        nargs=1,
        help="Action to perform: "
        "start -> start the given submitter, "
        "stop -> stop the given submitter if running, "
        "list -> list the currently running submitters"
        " from the given file,"
        " or any custom action defined for your"
        " submitter",
    )
    parser.add_argument(
        "submitter_file",
        type=str,
        help="Name of the Python module file containing the"
        " declaration for the Submitter to use",
    )
    # Optional arguments
    parser.add_argument(
        "-n",
        type=str,
        default=None,
        help="Name of the Submitter instance to use, if "
        "more than one is present in the given file. "
        "CAREFUL: this is the name of the variable, not the "
        "one passed as parameter in the constructor.",
    )
    parser.add_argument(
        "-nohup",
        action="store_true",
        default=False,
        help="If True, nohup is used to make sure that the "
        "process does not quit in case the user logs out "
        "(for example on a login node of an HPC machine).",
    )

    try:
        args = parser.parse_args()
    except ValueError:
        sys.exit(
            "Invalid submitter_file argument: file is not a valid Python" " module"
        )

    # First, check that the required submitter file exists and that it
    # contains what we need
    abs_subm_file = os.path.abspath(args.submitter_file)
    loaded_module = import_module(abs_subm_file)

    # Now load the actual Submitter!
    subms = {
        v: getattr(loaded_module, v)
        for v in dir(loaded_module)
        if v[:2] != "__" and issubclass(getattr(loaded_module, v).__class__, Submitter)
    }

    if len(subms) == 0:
        sys.exit("No Submitter instance found in " f"{args.submitter_file}")
    elif len(subms) > 1 and args.n is None:
        sys.exit(
            "Too many Submitter instances found in " f"{args.submitter_file}"
        )
    elif args.n is not None and args.n not in subms:
        sys.exit(
            f"Submitter of name {args.n} " f"not found in {args.submitter_file}"
        )
    # We're fine!
    if args.n is not None:
        submitter_name = args.n
    else:
        submitter_name = list(subms.keys())[0]

    if args.action[0] == "start":
        # Ok, first let's check that nothing like that is running already
        subm_l = Submitter.list()
        for s in subm_l:
            if os.path.abspath(s[0]) == abs_subm_file and submitter_name == s[1]:
                sys.exit("The requested submitter is already" " running")
        cmd = [
            "python",
            "-m",
            "soprano.hpc.submitter._spawn",
            abs_subm_file,
            submitter_name,
            "&",
        ]
        if args.nohup:
            cmd = ["nohup"] + cmd
        sp.Popen(cmd)
        print(
            "Submitter "
            f"{submitter_name} from file {args.submitter_file} started"
        )
    elif args.action[0] == "stop":
        # PKILL the process
        succ = Submitter.stop(abs_subm_file, submitter_name)
        if succ:
            print(
                "Submitter "
                f"{submitter_name} from file {args.submitter_file} stopped"
            )
        else:
            print("The requested submitter is not running")
    elif args.action[0] == "list":
        subm_l = Submitter.list()
        tabf = "{1: >10}\t| {2: >10}\t| {0: >10}"
        print(tabf.format("File", "Name", "Time"))
        for s in subm_l:
            if os.path.abspath(s[0]) != abs_subm_file:
                continue
            print(tabf.format(*s))
    elif args.action[0] in subms[submitter_name]._user_signals:
        signum, _ = subms[submitter_name]._user_signals[args.action[0]]
        subm_l = Submitter.list()
        for s in subm_l:
            if os.path.abspath(s[0]) == abs_subm_file and s[1] == submitter_name:
                os.kill(s[3], signum)
                break
    else:
        print(f"Unknown action {args.action[0]}")


submitter_handler()



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Definition of the Collection class.

It handles multiple Atoms ASE objects and mirrors in this sense the structure
of the Atoms object itself.
"""


import glob
import inspect
import os
import shutil
import sys
import warnings

import ase
import numpy as np

# 2-to-3 compatibility
try:
    import cPickle as pickle
except ImportError:
    import pickle
# More 2-to-3 compatibility, defining a wrapper function for inspect
# Internal imports
from ase import io as ase_io
from ase.build import niggli_reduce
from ase.calculators.singlepoint import SinglePointCalculator

from soprano import utils

utils.customize_warnings()


class _AllCaller:

    """_AllCaller class.

    A meta-object that serves the purpose of calling a function on all members
    of a list in a natural way.
    """

    def __init__(self, all_list, all_class=None):
        """Initialize the AllCaller with an 'all' list"""
        if all_class is None:
            self._class = all_list[0].__class__
        else:
            self._class = all_class
        if not all([x.__class__ is self._class for x in all_list]):
            raise ValueError(
                "Elements of list passed to an _AllCaller" " must be of the same type."
            )
        self._all = all_list
        # Now get a list of all members in common between these instances
        if len(self._all) > 0:
            self._instance_attrs = set.intersection(
                *[set(ins.__dict__.keys()) for ins in self._all]
            )

    def __getattr__(self, name):
        """Here's the magic of the class - when a method isn't found belonging
        to it, go looking for it in its ._all list..."""

        if hasattr(self._class, name):
            attr = getattr(self._class, name)
            # Is it a function?
            if not callable(attr):
                return np.array([getattr(x, name) for x in self._all])

            def iterfunc(*args, **kwargs):
                return np.array([getattr(x, name)(*args, **kwargs) for x in self._all])

            return iterfunc
        elif name in self._instance_attrs:
            # It's an instance attribute
            return np.array([getattr(x, name) for x in self._all])
        else:
            raise AttributeError(
                f"Not all '{self._class.__name__}' objects have attribute" f" '{name}'"
            )

    def map(self, f, *args, **kwargs):
        """Map a function to each element of the ._all list and return the
        results."""

        # First, check the signature
        if not callable(f):
            raise TypeError("Only functions can be mapped")

        nargs, nargs_def = utils.inspect_args(f)
        if nargs < 1 + len(args) + len(kwargs):
            # Function is invalid!
            raise ValueError("Invalid function passed to map")
        return [f(x, *args, **kwargs) for x in self._all]


class AtomsCollection:

    """AtomsCollection object.

    An AtomsCollection represents a group of ASE Atoms objects.
    It handles them together, can perform mass operations on them, and stores
    arrays of informations related to them.
    """

    def __init__(
        self,
        structures=[],
        info={},
        cell_reduce=False,
        progress=False,
        suppress_ase_warnings=True,
    ):
        """
        Initialize the AtomsCollection

        | Args:
        |    structures (list[str] or list[ase.Atoms]): list of file names or
        |                                               Atoms that will form
        |                                               the collection
        |    info (dict): dictionary of general information to attach
        |                 to this collection
        |    cell_reduce (bool): if True, perform a Niggli cell reduction on
        |                        all loaded structures
        |    progress (bool): visualize a progress bar for the loading process
        |    suppress_ase_warnings (bool): suppress annoying ASE warnings when
        |                                  loading files (default is True)
        """

        # Start by parsing out the structures
        self.structures = []

        if isinstance(structures, ase.Atoms):
            # Well, it's just one...
            structures = [structures]
        elif inspect.isgenerator(structures):
            # Let's unravel it
            iter_structs = structures
            structures = []
            for s in iter_structs:
                structures.append(s)

        if progress:
            sys.stdout.write("Loading collection...\n")
        s_n = len(structures)
        for s_i, struct in enumerate(structures):
            if progress:
                # Progress bar
                sys.stdout.write(f"\rLoading: {utils.progbar(s_i + 1, s_n)}")
            # Is it an Atoms object?
            if type(struct) is ase.Atoms:
                self.structures.append(ase.Atoms(struct))
                # Copy all arrays
                for k in struct.arrays.keys():
                    if not self.structures[-1].has(k):
                        self.structures[-1].new_array(k, struct.get_array(k))
                if struct.calc is not None:
                    # Prevents pointless attempts at re-calculating
                    self.structures[-1].calc._old_atoms = self.structures[-1]
            # Or is it a string?
            elif utils.is_string(struct):
                with utils.silence_stdio(suppress_ase_warnings, suppress_ase_warnings):
                    self.structures.append(ase_io.read(str(struct)))
                # If there's no name, give it the filename
                if "name" not in self.structures[-1].info:
                    self.structures[-1].info["name"] = utils.seedname(struct)
            else:
                raise TypeError(
                    "Structures must be Atoms objects or valid "
                    "file names,"
                    f" not {type(struct).__name__}"
                )
            if cell_reduce:
                # Here we must keep the energy if it was present
                # We do this by hand because ASE has its good reasons
                # for severing the atoms-calculator connection when changing
                # the unit cell.
                try:
                    _E = self.structures[-1].calc.results["energy"]
                except (KeyError, AttributeError):
                    _E = None
                niggli_reduce(self.structures[-1])
                if _E is not None:
                    _calc = SinglePointCalculator(self.structures[-1], energy=_E)
                    self.structures[-1].set_calculator(_calc)

        if progress:
            sys.stdout.write(f"\nLoaded {s_n} structures\n")

        self._all = _AllCaller(self.structures, ase.Atoms)

        self._arrays = {}

        # Now assign the info
        if type(info) is not dict:
            raise TypeError("Info must be dict," f" not {type(info).__name__}")
        else:
            self.info = info.copy()

    def __add__(self, other):
        """Addition of two collections brings a merging"""

        if not isinstance(other, self.__class__):
            raise TypeError(
                "'AtomsCollection' does not support operator +"
                " with object of type"
                f" '{type(other).__name__}'"
            )

        # Create a common collection, join arrays where present, copy all info
        all_struct = list(self.structures)
        all_struct += list(other.structures)

        all_info = dict(self.info)
        for k in other.info:
            if k not in all_info:
                all_info[k] = other.info[k]
            else:
                all_info.pop(k)

        sum_struct = AtomsCollection(all_struct, all_info)

        # Now arrays
        for aname in self._arrays:
            # Grab the shape
            shape = self._arrays[aname].shape[1:]
            # Check if present in the other as well
            all_arr = list(self.get_array(aname))
            if aname in other._arrays:
                all_arr += list(other.get_array(aname))
            else:
                all_arr += [np.zeros(shape) * np.nan] * other.length
            sum_struct.set_array(aname, all_arr, shape=shape)

        for aname in other._arrays:
            # Can not be present in both
            if aname in self._arrays:
                continue
            # Grab the shape
            shape = other._arrays[aname].shape[1:]
            all_arr = list(other.get_array(aname))
            all_arr = [np.zeros(shape)] * self.length + all_arr
            sum_struct.set_array(aname, all_arr, shape=shape)

        return sum_struct

    def __iadd__(self, other):
        self = self + other
        return self

    def __getitem__(self, indices):
        """Allow sophisticated slicing"""

        if type(indices) is int:
            # Special case, a single element!
            indices = indices % len(self)
            indices = slice(indices, indices + 1)

        try:
            struct_slice = self.structures[indices]
        except TypeError:
            indices = np.array(indices)
            if indices.dtype == bool:
                indices = np.where(indices)[0]  # Support for bool arrays
            struct_slice = [self.structures[i] for i in indices]

        sliced = AtomsCollection(struct_slice, info=self.info)

        # Now to add the arrays
        for a in self._arrays:
            sliced.set_array(a, self._arrays[a][indices])

        return sliced

    def __iter__(self):
        return self.structures.__iter__()

    def __deepcopy__(self, memodict={}):
        """Protects against problems with infinite recursion in AllCaller"""

        dcopy = AtomsCollection(self.structures, self.info)
        for arr in self._arrays:
            dcopy.set_array(arr, self._arrays[arr])

        return dcopy

    def __len__(self):
        return self.length

    @property
    def length(self):
        return len(self.structures)

    @property
    def all(self):
        return self._all

    def set_array(self, name, a, dtype=None, shape=None, args={}):
        """Add or modify an array of data related to the Atoms objects
        in this collection.

        | Args:
        |   name (str): name of the array to operate on.
        |   a (np.ndarray or function<Atoms, \\*\\*kwargs>
        |                    => Any): the data to assign to the array (must
        |                             be same length as the collection) or
        |                             a function that takes an Atoms object
        |                             as the first argument and returns a
        |                             value. This will be mapped over the
        |                             structures to create the array.
        |   dtype (type): type to cast the values of the array to.
        |   shape (tuple [int]): shape of each entry of the array. Will be
        |                        checked if provided.
        |   args (dict): named arguments to pass to the function provided
        |                as a. Will be ignored if an array is passed instead.

        """

        # a can be an actual array or a function that operates on each
        # separate Atoms object and returns a value

        a = np.array(a, dtype)
        if a.shape == ():
            a = a.item()
            if callable(a):
                # It's a function
                a = np.array(self.all.map(a, **args), dtype)
            else:
                # Invalid
                raise TypeError(
                    "new_array requires to pass either an array"
                    " or a function taking an Atoms object as its"
                    " first argument and returning a value."
                )

        # Now check that the shape is valid
        if shape is not None:
            if shape == (1,):
                targ_shape = (self.length,)
            else:
                targ_shape = (self.length,) + shape
            if a.shape != targ_shape:
                raise ValueError("Array of invalid shape passed to new_array")
        elif a.shape[0] != self.length:
            raise ValueError(
                "Array passed to new_array should be"
                " as long as the number of structures"
            )

        # And finally, assign
        self._arrays[name] = a

    def get_array(self, name, copy=True):
        """Get a copy of an array of given name (or a reference if copy=False)

        | Args:
        |   name (str): name of the array to retrieve.
        |   copy (bool): if the array should be copied or a reference should
        |                be returned instead.

        | Returns:
        |   array (np.ndarray): the requested array

        """

        if name not in self._arrays:
            raise ValueError(f"Array '{name}' does not exist")
        else:
            if copy:
                return np.array(self._arrays[name])
            else:
                # For NumPy 2.0 compatibility, use asarray to avoid copy when possible
                return np.asarray(self._arrays[name])

    def has(self, name):
        """Check if array of given name exists"""

        return name in self._arrays

    def set_calculators(self, calctype, labels=None, params={}):
        """Set an ASE calculator on each structure in the collection,
        and set said calculator's parameters.

        | Args:
        |   calctype (ASE Calculator type): the type of calculator
        |                                   to instantiate.
        |   labels (Optional[list[str]]): names to use for the calculators'
        |                                 files. If not present, random
        |                                 generated names are used.
        |   params (Optional[dict]): parameters of the calculator to set.

        """

        # First, a check
        from ase.calculators.calculator import Calculator as cCalculator
        from ase.calculators.calculator import FileIOCalculator as ioCalculator
        
        # Check if calctype is a subclass of ASE Calculator
        # In newer ASE versions, ase.calculators.general doesn't exist
        try:
            from ase.calculators.general import Calculator as gCalculator
            valid_bases = (gCalculator, cCalculator, ioCalculator)
        except ImportError:
            # Fallback for newer ASE versions
            valid_bases = (cCalculator, ioCalculator)

        if not any(issubclass(calctype, base) for base in valid_bases):
            raise TypeError("calctype must be a type of ASE Calculator")

        if labels is not None and len(labels) != self.length:
            raise ValueError("labels must be long as the collection itself")

        # Then set it up
        for i, s in enumerate(self.structures):
            if labels is None:
                # First: do we have a name?
                if "name" in s.info:
                    label = s.info["name"]
                else:
                    label = f"struct_{i}"
            else:
                label = labels[i]
            calc = calctype(atoms=s, label=str(label), **params)
            # To make sure...
            s.calc = calc

    def run_calculators(self, properties=None, system_changes=None):
        """Run all previously set ASE calculators.

        | Args:
        |   properties (list[str]): list of properties to calculate (depends
        |                           on type of Calculator used)
        |   system_changes (list[str]): list of changes to the structure
        |                               since the last calculation. Can be
        |                               any combination of these five:
        |                               'positions', 'numbers', 'cell',
        |                               'pbc', 'initial_charges' and
        |                               'initial_magmoms'.

        """

        # First, check if we even have those
        if any([c is None for c in self.all.calc]):
            raise RuntimeError("Not all structures in collection" " have a calculator")
        kwargs = {}
        if properties is not None:
            kwargs["properties"] = properties
        if system_changes is not None:
            kwargs["system_changes"] = system_changes
        self.all.map(lambda s: s.calc.calculate(atoms=s, **kwargs))

    def chunkify(self, chunk_size=None, chunk_n=None):
        """Split this collection into multiple collections based on either
        size or number of chunks.

        | Args:
        |   chunk_size (Optional[int]): maximum size of a generated chunk
        |   chunk_n (Optional[int]): number of chunks to generate

        | Returns:
        |   chunks (list[AtomsCollection]): a list of the generated chunks

        """

        if [chunk_size, chunk_n].count(None) != 1:
            raise RuntimeError(
                "Only one between chunk_size and chunk_n" "must be passed"
            )

        if chunk_size is not None and type(chunk_size) is not int:
            raise TypeError("chunk_size must be an int")
        if chunk_n is not None and type(chunk_n) is not int:
            raise TypeError("chunk_n must be an int")

        # Now determine the size
        if chunk_size is None:
            chunk_size = int(np.ceil(self.length * 1.0 / chunk_n))

        struct_chunks = [
            self.structures[i : i + chunk_size]
            for i in range(0, self.length, chunk_size)
        ]
        chunks = [AtomsCollection(s, self.info) for s in struct_chunks]

        # Now the arrays
        for aname in self._arrays:
            for i, c in enumerate(chunks):
                c.set_array(
                    aname, self._arrays[aname][i * chunk_size : i * (chunk_size + 1)]
                )

        return chunks

    def sorted_byarray(self, name, reverse=False):
        """Return a copy of this collection sorted by a given array.

        | Args:
        |   name (str): name of the array to use for the sorting
        |   reverse (Optional[bool]): reverse order of sorting (max to min)

        | Returns:
        |   sorted (AtomsCollection): a sorted copy of the collection

        """

        # First, check that we do have the array
        if not self.has(name):
            raise ValueError(f"Array '{name}' does not exist")

        arr_names = list(self._arrays.keys())
        data_block = zip(self.structures, *[self._arrays[n] for n in arr_names])
        key_i = arr_names.index(name)
        data_block = list(
            zip(*sorted(data_block, key=lambda x: x[key_i + 1], reverse=reverse))
        )

        sorted_copy = AtomsCollection(data_block[0], info=self.info)
        for ai, an in enumerate(arr_names):
            sorted_copy.set_array(an, data_block[ai + 1])

        return sorted_copy

    def filter(self, filter_func):
        """Return a collection composed only of the elements for which a given
        filter function returns True.

        | Args:
        |   filter_func (function<Atoms>
        |                       => bool): filter function. Should take an
        |                                 Atoms object and return a boolean

        | Returns:
        |   filtered (AtomsCollection): the filtered version of the collection

        """

        filter_slice = []

        for i, s in enumerate(self.structures):
            if filter_func(s):
                filter_slice.append(i)

        return self[filter_slice]

    def classify(self, classes):
        """Return a dictionary of collections based on the names of assigned
        classes.

        | Args:
        |   classes (np.ndarray): array of the class to which each structure
        |                         belongs. For example [1, 2, 1] will put the
        |                         first and third structures in class 1 and
        |                         the other in class 2. The classes can be any
        |                         hashable types, like int or str.

        | Returns:
        |   classified (dict): a dictionary using class names as keys and
        |                      sliced collections as values

        """

        classes = np.array(classes)
        classified = {k: self[np.where(classes == k)[0]] for k in set(classes)}

        return classified

    def save(self, filename):
        """Simply save a pickled copy to a given file path"""

        # Pickling doesn't deal well with the _AllCaller, so we get rid of it
        selfcopy = self[:]
        selfcopy._all = None

        with open(filename, "wb") as f:
            pickle.dump(selfcopy, f, protocol=2)

    @staticmethod
    def load(filename):
        """Load a pickled copy from a given file path"""

        with open(filename, "rb") as f:
            coll = pickle.load(f)

        if not isinstance(coll, AtomsCollection):
            raise ValueError("File does not contain an AtomsCollection" " object")
        # Restore the _AllCaller
        coll._all = _AllCaller(coll.structures, ase.Atoms)
        return coll

    @staticmethod
    def check_tree(path):
        """Checks if a path is a valid 'tree' format for a collection. This is
        any folder that satisfies the following conditions:

        - contains a .collection file storing metadata
        - contains a series of folders matching the list stored in the
          .collection file, and nothing else

        This function will return 0 if both conditions are satisfied, 1 if
        only the first is, 2 if no .collection file is found, and -1 if the
        folder itself doesn't exist.

        | Args:
        |   path (str): path to check for whether it matches or not the
        |               collection pattern

        | Returns:
        |   result (int): 0, 1 or 2 depending on the outcome of the checks
        """

        if not os.path.exists(path):
            return -1

        # Begin by checking whether there is a .collection file
        try:
            with open(os.path.join(path, ".collection"), "rb") as f:
                coll = pickle.load(f)
        except (OSError, UnicodeDecodeError):
            return 2  # No or invalid .collection file found

        # Check if the directories match
        dirlist = coll["dirlist"]
        dirs = glob.glob(os.path.join(path, "*"))
        all_dirs = all([os.path.isdir(d) for d in dirs])
        dirs = [os.path.relpath(d, path) for d in dirs]

        # Are they even all directories?
        if not all_dirs or set(dirlist) != set(dirs):
            return 1

        return 0

    def save_tree(
        self,
        path,
        save_format,
        name_root="structure",
        opt_args={},
        safety_check=3,
        suppress_ase_warnings=True,
    ):
        """Save the collection's structures as a series of folders, named like
        the structures, inside a given parent folder (that will be created if
        not present). Arrays and info are stored in a pickled .collection file
        which works as metadata for the whole directory tree.
        The files can be saved in a format of choice, or a function can be
        passed that will save them in a custom way. Only one collection can be
        saved per folder.

        | Args:
        |   path (str): folder path in which the collection should be saved.
        |   save_format (str or function): format in which the structures
        |                                  should be saved.
        |                                  If a string, it will be used as a
        |                                  file extension. If a function, it
        |                                  must take as arguments the
        |                                  structure (an ase.Atoms object)
        |                                  the save path (a string), and any
        |                                  additional arguments passed as
        |                                  opt_args, and take care of saving
        |                                  the required files.
        |   name_root (str): name prefix to be used for structures when a name
        |                    is not available in their info dictionary
        |   opt_args (dict): dictionary of additional arguments to pass to
        |                    either ase.io.write (if save_format is a string)
        |                    or to the save_format function.
        |   safety_check (int): how much care should be taken not to overwrite
        |                       potentially important data in path. Can be a
        |                       number from 0 to 3.
        |                       Here's the meaning of the codes:
        |
        |                       3 (default): always ask before overwriting an
        |                         existing folder that passes the check_tree
        |                         control, raise an exception otherwise;
        |                       2: overwite any folder that passes fully the
        |                          check_tree control, raise an exception
        |                          otherwise;
        |                       1: overwrite any folder that passes fully the
        |                          check_tree control, ask for user input
        |                          otherwise;
        |                       0 (DANGER - use at your own risk!): no checks,
        |                         always overwrite path.

        """

        check = AtomsCollection.check_tree(path)

        def ow_ask(path):
            return (
                utils.safe_input(
                    f"Folder {path} exists, " "overwrite (y/n)?"
                ).lower()
                == "y"
            )

        if check > -1:
            # The folder exists
            if check == 0:
                if safety_check >= 3:
                    # Ask for permission
                    perm = ow_ask(path)
                else:
                    perm = True
            elif safety_check >= 2:
                raise OSError(

                        f"Trying to overwrite folder {path} which did"
                        f" not pass check_tree control (result {check})"
                        " with safety_check level "
                        f"{safety_check}"

                )
            elif safety_check == 1:
                perm = ow_ask(path)
            else:
                perm = True

            if not perm:
                print(f"Can not overwrite folder {path}, skipping...")

            shutil.rmtree(path)

        # Re-create folder
        os.mkdir(path)

        # Format type?
        is_ext = utils.is_string(save_format)
        is_func = callable(save_format)
        if not (is_ext or is_func):
            raise ValueError("Invalid save_format passed to save_tree")

        dirlist = []
        for i, s in enumerate(self.structures):
            sname = s.info.get("name", f"{name_root}_{i + 1}")
            fold = os.path.join(path, sname)
            try:
                os.mkdir(fold)
            except OSError:
                shutil.rmtree(fold)
                os.mkdir(fold)
            with utils.silence_stdio(suppress_ase_warnings, suppress_ase_warnings):
                if is_ext:
                    ase_io.write(
                        os.path.join(fold, sname + "." + save_format), s, **opt_args
                    )
                elif is_func:
                    save_format(s, fold, **opt_args)

            dirlist.append(sname)

        with open(os.path.join(path, ".collection"), "wb") as f:
            pickle.dump(
                {"dirlist": dirlist, "arrays": self._arrays, "info": self.info},
                f,
                protocol=2,
            )

    @staticmethod
    def load_tree(
        path,
        load_format,
        opt_args={},
        safety_check=3,
        tolerant=False,
        suppress_ase_warnings=True,
    ):
        """Load a collection's structures from a series of folders, named like
        the structures, inside a given parent folder, as created by save_tree.
        The files can be loaded from a format of choice, or a
        function can be passed that will load them in a custom way.

        | Args:
        |   path (str): folder path in which the collection should be saved.
        |   load_format (str or function): format from which the structures
        |                                  should be loaded.
        |                                  If a string, it will be used as a
        |                                  file extension. If a function, it
        |                                  must take as arguments the load
        |                                  path (a string) and any additional
        |                                  arguments passed as opt_args, and
        |                                  return the loaded structure as an
        |                                  ase.Atoms object.
        |   opt_args(dict): dictionary of additional arguments to pass to
        |                   either ase.io.read (if load_format is a string)
        |                   or to the load_format function.
        |   safety_check (int): how much care should be taken to verify the
        |                       folder that is being loaded. Can be a number
        |                       from 0 to 3.
        |                       Here's the meaning of the codes:
        |
        |                       3 (default): only load a folder if it passes
        |                         fully the check_tree control;
        |                       2: load any folder that has a valid
        |                          .collection file, but only the listed
        |                          subfolders;
        |                       1: load any folder that has a valid
        |                          .collection file, all subfolders. Array
        |                          data will be discarded;
        |                       0: no checks, try to load from all subfolders.
        |   tolerant (bool): if set to true, proceeds to load the
        |                    structures into an AtomsCollection, even
        |                    if some of the structures could not be
        |                    read.

        | Returns:
        |   coll (AtomsCollection): loaded collection

        """

        check = AtomsCollection.check_tree(path)

        if check == -1:
            raise OSError(f"Folder {path} does not exist")

        dirlist = []
        if check < 2:
            with open(os.path.join(path, ".collection"), "rb") as f:
                coll = pickle.load(f)
            if check == 1 and safety_check == 3:
                raise OSError(
                    f"Folder {path} is not a valid collection " "tree"
                )
            if safety_check >= 2:
                dirlist = coll["dirlist"]
            else:
                dirlist = [
                    os.path.relpath(d, path)
                    for d in glob.glob(os.path.join(path, "*"))
                    if os.path.isdir(d)
                ]
        else:
            if safety_check > 0:
                raise OSError(
                    f"Folder {path} is not a valid collection " "tree"
                )
            dirlist = [
                os.path.relpath(d, path)
                for d in glob.glob(os.path.join(path, "*"))
                if os.path.isdir(d)
            ]

        # Format type?
        is_ext = utils.is_string(load_format)
        is_func = callable(load_format)
        if not (is_ext or is_func):
            raise ValueError("Invalid load_format passed to load_tree")

        structures = []
        for d in dirlist:
            try:
                if is_ext:
                    with utils.silence_stdio(
                        suppress_ase_warnings, suppress_ase_warnings
                    ):
                        s = ase_io.read(
                            os.path.join(path, d, d + "." + load_format), **opt_args
                        )
                elif is_func:
                    s = load_format(os.path.join(path, d), **opt_args)
                structures.append(s)
            except Exception as e:
                warnings.warn(str(e))

        if check < 2:
            info = coll["info"]
        else:
            info = {}

        percentage_failed = (1 - len(structures) / len(dirlist)) * 100

        if percentage_failed > 0:

            if percentage_failed == 100:
                raise OSError(
                    f"{percentage_failed:.2f}% of structures could not be loaded."
                )
                return
            elif not tolerant:
                raise OSError(
                    f"{percentage_failed:.2f}% of structures could not be loaded. Set"
                    " tolerant to True if you would still"
                    " like to load the remaining structures."
                )
                return
            else:
                percentage_success = 100 - percentage_failed
                info["percentage_loaded"] = percentage_success

                warnings.warn(
                    f"{percentage_failed:.2f}% of structures could not be loaded."
                )
        else:
            print("100% of structures loaded successfully.")

        loaded_coll = AtomsCollection(structures, info=info)

        if safety_check >= 2:
            arrays = coll["arrays"]
            for k, a in arrays.items():
                loaded_coll.set_array(k, a)

        return loaded_coll


if __name__ == "__main__":

    # Just load whatever was passed in the command line!

    if len(sys.argv) > 1:
        coll = AtomsCollection(sys.argv[1:], progress=True, cell_reduce=True)



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""CLI to plot NMR results from .magres file(s).
TODOs:
- [ ] add stylesheets
- [ ] add dipolar/j-coupling scaling of markers
    - [ ] Basic version done, but maybe check the assumptions made when running with the reduce or average groups options...
- [ ] 1D plots -- basic simulation of 1D NMR?
"""

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "May 09, 2023"


import logging

import click
import click_log
import matplotlib.pyplot as plt
import numpy as np

from soprano.calculate.nmr import NMRCalculator
from soprano.calculate.nmr.nmr import NMRData2D, NMRPlot2D, PlotSettings
from soprano.properties.nmr import *
from soprano.properties.nmr import MSIsotropy
from soprano.scripts.cli_utils import PLOT_OPTIONS, add_options, viewimages
from soprano.scripts.nmr import nmr_extract_multi, print_results
from soprano.selection import AtomSelection

# logging
logging.captureWarnings(True)
logger = logging.getLogger("cli")
click_log.basic_config(logger)


@click.command()
@click.argument("files", nargs=-1, type=click.Path(exists=True))
@add_options(PLOT_OPTIONS)
def plotnmr(
    files,
    isotopes,
    average_group,
    euler_convention,
    references,
    gradients,
    subset,
    reduce,
    query,
    plot_type,
    x_element,
    y_element,
    yaxis_order,
    rcut,
    xlim,
    ylim,
    show_markers,
    marker_symbol,
    scale_marker_by,
    max_marker_size,
    marker_color,
    marker_linewidth,
    show_marker_legend,
    show_diagonal,
    show_grid,
    show_connectors,
    show_ticklabels,
    show_heatmap,
    xbroadening,
    ybroadening,
    colormap,
    show_contour,
    contour_levels,
    contour_color,
    contour_linewidth,
    plot_filename,
    plot_shielding,  ## force-plot the shielding even if references are given
    verbosity,
    symprec,
    precision,
    view,
):
    """
    Plot the NMR spectrum from a .magres file.
    """
    # Hard-code a few options for nmr_extract since the user doesn't need to specify them
    properties = ["ms"]  # we at least need to extract the MS
    include = ["MS_shielding"]  # we only need the MS columns for plotting
    if references:
        include.append("MS_shift")
    exclude = None
    merge = True  # if multiple files are given, we have to merge them
    sortby = None
    sort_order = "ascending"
    combine_rule = "mean"

    # set verbosity
    if verbosity == 0:
        logger.setLevel(logging.WARNING)
    elif verbosity == 1:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.DEBUG)

    dfs, images = nmr_extract_multi(
        files,
        subset=subset,
        merge=merge,
        isotopes=isotopes,
        references=references,
        gradients=gradients,
        reduce=reduce,
        average_group=average_group,
        symprec=symprec,
        properties=properties,
        euler_convention=euler_convention,
        sortby=sortby,
        sort_order=sort_order,
        include=include,
        exclude=exclude,
        query=query,
        logger=logger,
    )

    if view:
        viewimages(images)

    # write to file(s)
    if verbosity > 0:
        print_results(dfs)
    if len(dfs) > 1:
        logger.warning(
            "More than one dataframe extracted. Only plotting the first one."
        )
    if len(dfs) == 0:
        logger.error("No dataframes extracted. Aborting.")
        return 1
    atoms = images[0]

    if plot_type == "2D":
        if not y_element:
            y_element = x_element

        shift = not plot_shielding if plot_shielding is not None else references != {}

        # Create NMRData2D instance
        nmr_data = NMRData2D(
            atoms=atoms,
            xelement=x_element,
            yelement=y_element,
            rcut=rcut,
            references=references,
            gradients=gradients,
            isotopes=isotopes,
            is_shift=shift,
            include_quadrupolar=False,
            yaxis_order=yaxis_order,
            correlation_strength_metric=scale_marker_by,
        )

        # Define plot settings
        plot_settings = PlotSettings(
            xlim=xlim,
            ylim=ylim,
            show_markers=show_markers,
            marker=marker_symbol,
            max_marker_size=max_marker_size,
            marker_linewidth=marker_linewidth,
            plot_filename=plot_filename,
            show_lines=show_grid,
            show_diagonal=show_diagonal,
            show_connectors=show_connectors,
            show_labels=show_ticklabels,
            show_heatmap=show_heatmap,
            show_contour=show_contour,
            colormap=colormap,
            marker_color=marker_color,
            show_legend=show_marker_legend,
            contour_levels=contour_levels,
            contour_color=contour_color,
            contour_linewidth=contour_linewidth,
            x_broadening=xbroadening,
            y_broadening=ybroadening,

        )

        # Create NMRPlot2D instance
        nmr_plot = NMRPlot2D(
            nmr_data=nmr_data,
            plot_settings=plot_settings,
        )

        # Generate the plot
        fig, ax = nmr_plot.plot()
        # if the user doesn't give an output file name, show the plot using the default matplotlib backend
        if not plot_filename:
            plt.show()
    elif plot_type == "1D":
        shift = not plot_shielding if plot_shielding is not None else references != {}
        sel = AtomSelection.all(atoms)
        element_sel = AtomSelection.from_element(atoms, x_element)
        sel = sel * element_sel
        atoms = sel.subset(atoms)
        # get the NMR calculator
        calc = NMRCalculator(atoms)
        if shift:
            logger.info(f"Setting references: {references}")
            calc.set_reference(ref=references[x_element], element=x_element)
            use_reference = True
            xlabel = f"{x_element} shift (ppm)"
        else:
            xlabel = f"{x_element} shielding (ppm)"
            use_reference = False

        if isotopes:
            calc.set_isotopes(isotopes)
        calc.set_powder(N=8)

        iso = MSIsotropy.get(atoms)
        max_iso = np.max(iso)
        min_iso = np.min(iso)
        iso_range = max_iso - min_iso
        max_iso += iso_range * 0.1
        min_iso -= iso_range * 0.1
        # get 1D plot data
        spec, freq = calc.spectrum_1d(
            x_element,
            min_freq=min_iso,
            max_freq=max_iso,
            bins=1001,
            freq_broad=0.05,
            freq_units="ppm",
            # effects=NMRFlags.Q_1_ORIENT,
            use_central=True,
            use_reference=use_reference,
        )
        # plot
        fig, ax = plt.subplots()
        ax.plot(freq, spec)

        ax.set_xlabel(xlabel)
        ax.set_ylabel("Intensity")
        if xlim:
            ax.set_xlim(xlim)
        if ylim:
            ax.set_ylim(ylim)
        if use_reference:
            ax.invert_xaxis()
        if plot_filename:
            fig.savefig(plot_filename)
        else:
            plt.show()
    else:
        logger.error("Invalid plot type. Aborting.")

    return 0



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import warnings

import numpy as np
from ase.geometry import get_distances
from scipy.optimize import linear_sum_assignment

from soprano.properties import AtomsProperty


class RemapIndices(AtomsProperty):

    """
    RemapIndices

    Returns a list of the indices which will remap a given structure to the
    one given as reference. Remapping means creating a one-to-one correspondence
    between atoms based on their distances (which may include the periodic
    boundaries).

    | Parameters:
    |   reference (ase.Atoms):  Reference structure to map to. Required
    |   mic (bool):             If True, take into account periodic boundaries
    |                           via the minimum image convention. We use the general method
    |                           from ASE. It's slow in some cases but robust. 
    |                           Default is True.
    |   check_species (bool):   If True, only compare atoms of the same species. 
    |                           Default is True.
    |   tolerance (float):      Tolerance (in Angstroms) for the distance comparison. If no atom
    |                           is found within this distance, an error is raised.
    |                           Default is 0.1 Angstrom.

    | Returns:
    |   remap_indices ([int]):  List of indices for the structure in the order
    |                           in which they will make it map to the reference
    |                           best 
    """

    default_name = "remap_indices"
    default_params= {
        "reference": None,
        "mic": True,
        "check_species": True,
        "tolerance": 0.1}




    @staticmethod
    def extract(s, reference, mic, check_species, tolerance):

        # First, are these even compatible structures?
        f1 = s.get_chemical_formula()
        f2 = reference.get_chemical_formula()
        if f1 != f2:
            raise ValueError('Structures do not have the same formula')

        n = len(s)
        # Check species
        # group indices by species so that we only compare the same species' positions
        if check_species:
            species_list = list(set(s.get_chemical_symbols()))
            s_species_groups = [[atom.index for atom in s if atom.symbol in species] for species in species_list]
            r_species_groups = [[atom.index for atom in reference if atom.symbol in species] for species in species_list]
        else:
            species_list = ['all']
            s_species_groups = [range(n)]
            r_species_groups = [range(n)]

        # Now, loop over the species groups
        assignments = []
        for s_species_group, r_species_group, species in zip(s_species_groups, r_species_groups, species_list):

            # Start by computing the distance matrix
            p1 = reference[r_species_group].positions
            p2 = s[s_species_group].positions

            # use the minimum image convention
            if mic:
                pbc = reference.get_pbc()
                cell = reference.cell
                rmat, dmat = get_distances(p1, p2=p2, cell=cell, pbc=pbc)
            else:
                rmat = p1[:, None]-p2[None, :]
                dmat = np.linalg.norm(rmat, axis=2)

            # Solve the linear sum assignment problem.
            row_ind, col_ind = linear_sum_assignment(dmat)

            # Check that the assigned atoms are within the tolerance
            if (dmat[row_ind, col_ind] > tolerance).any():
                # Some atoms are not within the tolerance. raise error
                # indices of atoms that don't match:
                bad_indices = np.where(dmat[row_ind, col_ind] > tolerance)
                raise ValueError(f'Atoms with indices: {bad_indices} are not within the tolerance distance of another atom')
            # append species-specific assignment
            assignments.append([s_species_group[i] for i in col_ind])

        # flatten the list of assignments
        new_indices = [item for sublist in assignments for item in sublist]
        # fix the species-sublisted order -> global order
        global_order = np.argsort([item for sublist in r_species_groups for item in sublist])
        new_indices = [new_indices[i] for i in global_order]

        if not all(reference.symbols == s.symbols[new_indices]):
            warnings.warn('Warning: s and reference have some atoms whose species do not match, \n'
            'Set check_species=True if you want to only match atoms of same species')
        return new_indices

class Remap(AtomsProperty):
    """
    Remap

    Returns a structure remapped to the one given as reference. 
    Remapping means creating a one-to-one correspondence between atoms based
    on their distances (which may include the periodic boundaries).

    | Parameters:
    |   reference (ase.Atoms):  Reference structure to map to. Required
    |   mic (bool):             If True, take into account periodic boundaries
    |                           via the minimum image convention. We use the general method
    |                           from ASE. It's slow but robust. 
    |                           Default is True.
    |   check_species (bool):   If True, only compare atoms of the same species. 
    |                           Default is True.
    |   tolerance (float):      Tolerance for the distance comparison. 
    |                           Default is 0.1 Angstrom.
    | Returns:
    |   remap_indices ([int]):  List of indices for the structure in the order
    |                           in which they will make it map to the reference
    |                           best 
    """

    default_name = "remap"
    default_params = {
        "reference": None,
        "mic": True,
        "check_species": True,
        "tolerance": 0.1}

    @staticmethod
    def extract(s, reference, mic, check_species, tolerance):

        indices = RemapIndices.extract(s, reference, mic, check_species, tolerance)

        return s[indices]



# Soprano - a library to crack crystals! by Simone Sturniolo
# Copyright (C) 2016 - Science and Technology Facility Council

# Soprano is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# Soprano is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.

# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
CLI to extract all molecules in a molecular crystal (in any ASE-readable format) and output the in individual structure files.

"""

__author__ = "J. Kane Shenton"
__maintainer__ = "J. Kane Shenton"
__email__ = "kane.shenton@stfc.ac.uk"
__date__ = "Dec. 13, 2023"


import logging
import os
from pathlib import Path

import click
import click_log
import numpy as np
from ase import Atoms, io
from ase.data import atomic_numbers

from soprano.data import build_custom_vdw
from soprano.properties.linkage import Molecules
from soprano.scripts.cli_utils import (
    add_options,
    has_CH_bonds,
    keyvalue_parser,
    view,
    viewimages,
)
from soprano.utils import has_cif_labels

# logging
logging.captureWarnings(True)
logger = logging.getLogger("cli")
click_log.basic_config(logger)

HEADER = """
###############################################
Extracting molecules from a molecular crystal
###############################################
"""

def validate_cell(ctx, param, value):
    if value is not None:
        parts = value.split()
        num_parts = len(parts)
        if num_parts not in (1, 3, 6, 9):
            raise click.BadParameter("Cell must have 1, 3, 6, or 9 values.")
        try:
            # Convert all parts to floats
            parts = np.array([float(part) for part in parts])
        except ValueError:
            raise click.BadParameter("All cell values must be valid floats.")
        # If num_parts is 9, reshape to 3x3 matrix
        if num_parts == 9:
            parts = parts.reshape((3, 3))
        if num_parts == 1:
            parts = np.eye(3) * parts[0]
        return parts
    return value

# options for the command line
# TODO - later maybe move to the cli_utils file


@click.command()
@click.argument("filename", type=click.Path(exists=True))
@click.option(
    "--seedname",
    "-s",
    type=click.STRING,
    default=None,
    help="Seedname for the output files. Defaults to the input filename without its extension.",
)
# output directory - create if it doesn't exist
@click.option(
    "--output-dir",
    "-o",
    type=click.Path(),
    default=".",
    help="Output directory for the extracted molecules.",
)
@click.option(
    "--format",
    "-f",
    type=click.STRING,
    default="xyz",
    help="Output file format for the extracted molecules. This can be any format supported by ASE.",
)
# disable file output
@click.option(
    "--no-write",
    is_flag=True,
    help="Disable output of the extracted molecules to files.",
)
# redefine unit cell
@click.option(
    "--cell",
    type=click.STRING,
    callback=validate_cell,
    default=None,
    help="Redefined unit cell for the extracted molecules. Use double-quotes to pass this option in. "
    "For example: ```--cell \"10.0 10.0 15.0\"```. "
    "This can be provides as a single float (-> a cubic cell is created with that lattice constant) or "
    "as three floats (`a b c` -> an orthorhombic cell is created with the given side lengths) or "
    "as six floats (`a b c alpha beta gamma`) or "
    "as nine floats (`a11 a12 a13 a21 a22 a23 a31 a32 a33`) for the full unit cell matrix. "
    "Units are in Angstroms."
)
# center molecules in the cell
@click.option(
    "--center",
    "--centre", # British spelling allowed too...
    "-c",
    is_flag=True,
    help="Center the extracted molecules in the unit cell.",
)
# enforce minimum vacuum
@click.option(
    "--vacuum",
    "--vac",
    type=click.FLOAT,
    default=None,
    help="Enforce a minimum vacuum between the extracted molecules and the cell boundaries. Applied in all directions."
    "Only applies if the `--center` flag is set."
    "See the ASE Atoms.center() documentation for how this works.",
)
# use cell-indices: yes/no?
@click.option(
    "--cell-indices/--no-cell-indices",
    is_flag=True,
    default=True,
    help="Use cell indices when outputting the individual structures. This is useful for molecules that are split across periodic boundaries."
    "default is to use the cell-indices.",
)
# -- Soprano Molecules kwargs:
@click.option(
    "--vdw-set",
    type=click.Choice(["ase", "jmol", "csd"]),
    default="csd",
    help="Set of Van der Waals radii to use. Default is csd S. Alvarez, 2013 (`<https://doi.org/10.1039/C3DT50599E>`_).",
)
@click.option(
    "--vdw-scale",
    type=click.FLOAT,
    default=1.0,
    help="Scaling factor to apply to the base Van der Waals radii values. Values bigger than one make for more tolerant molecules.",
)
@click.option(
    "--default-vdw",
    type=click.FLOAT,
    default=2.0,
    help="Default Van der Waals radius for species for whom no data is available. Default is 2.0 Angstroms.",
)
@click.option(
    "--vdw-custom",
    type=click.STRING,
    callback=keyvalue_parser,
    default="",
    help="A comma-separated list of custom Van der Waals radii to use, overriding the existing ones, expressed as: H:1,C:2 etc. Units are in Angstroms.",
)
# verbosity flag e.g. -v -vv -vvv
@click.option(
    "--verbosity",
    "-v",
    count=True,
    help="Increase verbosity. "
    "Use -v for info, -vv for debug. Defaults to only showing warnings and errors.",
)
@add_options([view])
def splitmols(
    filename,
    seedname,
    output_dir,
    format,
    no_write,
    view,
    cell,
    center,
    vacuum,
    cell_indices,
    vdw_set,
    vdw_scale,
    default_vdw,
    vdw_custom,
    verbosity,
):
    """

    Extract all molecules in a molecular crystal (in any ASE-readable format) and output the in individual structure files.

    """

    # Set up logging
    if verbosity == 0:
        logging.basicConfig(level=logging.WARNING)
    elif verbosity == 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.DEBUG)

    # Load the structure
    atoms = io.read(filename, index="-1")

    # works better if first wrapped to unit cell
    # atoms.wrap()

    if not has_CH_bonds(atoms):
        logger.warning(
            "No C-H bonds found in the structure. Are you sure this is a molecular crystal?"
        )

    # log the chosen vdW radii
    vdw_r = build_custom_vdw(vdw_set, vdw_scale, default_vdw, vdw_custom)
    elements = list(set(atoms.get_chemical_symbols()))
    logger.debug(f"Elements in the structure: {elements}")
    logger.debug("Using Van der Waals radii for the bond search:")
    for el in elements:
        logger.debug(f"{el}: {vdw_r[atomic_numbers[el]]}")

    # split into molecules
    molecules = extract_molecules(
        atoms,
        cell_indices,
        vdw_set=vdw_set,
        vdw_scale=vdw_scale,
        default_vdw=default_vdw,
        vdw_custom=vdw_custom,
    )

    Nmols = len(molecules)
    if Nmols == 0:
        err_msg = "No molecules found in the structure"
        logger.error(err_msg)
        raise RuntimeError(err_msg)

    # log
    logger.info(f"Found {Nmols} molecules")
    for i, mol in enumerate(molecules):
        logger.debug(f"Molecule {i}: {mol.get_chemical_formula()}")

    # redefine unit cell if required
    for i in range(Nmols):
        molecules[i] = redefine_unit_cell(molecules[i], cell, center, vacuum)

    # View the molecules if required
    if view:
        viewimages(molecules, reload_as_molecular=False) # disable reloading as molecular as this is already done

    # write the molecules to individual files
    if seedname is None:
        seedname = Path(filename).stem
    if not no_write:
        # Create output directory using pathlib
        output_path = Path(output_dir)
        output_path.mkdir(exist_ok=True, parents=True)
        write_molecules(molecules, output_dir, format, seedname)



def write_molecules(molecules, output_dir, format, seedname):
    """
    Write the molecules to individual files.
    """
    Nmols = len(molecules)
    # how many leading zeros do we need?
    Nzeros = int(np.log10(Nmols)) + 1
    for i, mol in enumerate(molecules):
        # make the filename
        filename = f"{seedname}_{i:0{Nzeros}}.{format}"
        logger.info(f"Writing molecule {i} to {filename}")
        # TODO think about what info to include in the file
        # make a copy of the atoms object without the arrays
        mol_temp = Atoms(
            mol.get_chemical_symbols(),
            positions=mol.get_positions(),
            cell=mol.get_cell(),
            pbc=mol.get_pbc(),
        )
        # include labels
        if has_cif_labels(mol):
            mol_temp.set_array("labels", mol.get_array("labels"))

        mol_temp.write(
            str(Path(output_dir) / filename),
            # format=format,
            # TODO - maybe add some more options here?
        )

def extract_molecules(atoms, use_cell_indices, **kwargs):
    mols = Molecules.get(atoms, **kwargs)

    molecules = []
    for mol in mols:
        molecules.append(mol.subset(atoms, use_cell_indices=use_cell_indices))

    return molecules


def redefine_unit_cell(atoms: Atoms,
                       cell: np.ndarray,
                       center=False,
                       vacuum=None):
    """
    Redefine the unit cell of the structure.
    If center is True, the structure is centered in the cell.
    (Centers the atoms in the unit cell, so there is the same
        amount of vacuum on all sides.)
    If vacuum is not None, a minimum vacuum is enforced.
    cell can be a single float (-> a cubic cell is created with that lattice constant) or
    as three floats (`a b c` -> an orthorhombic cell is created with the given side lengths) or
    as six floats (`a b c alpha beta gamma`) or
    a 3x3 np.array defining the full unit cell.
    """
    if cell is not None:
        atoms.set_cell(cell)
        atoms.set_pbc(True)

    if not center and vacuum:
        # TODO: handle these cases - maybe translating COM back to original position after centering + vacuum?
        err_msg = "Cannot enforce a minimum vacuum without centering the molecule. Please set the --center flag."
        logger.error(err_msg)
        raise RuntimeError(err_msg)

    if center:
        logger.debug("Centering the atoms in the cell")
        if vacuum:
            logger.debug(f"Enforcing a minimum vacuum of {vacuum} Å")
        atoms.center(vacuum=vacuum)
    return atoms
